from typing import TYPE_CHECKING

from django.db import models

from django_napse.core.models.wallets.wallet import Wallet

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import ExchangeAccount
    from django_napse.core.models.accounts.space import Space


class ConnectionWallet(Wallet):
    """A Wallet owned by a Connection."""

    owner = models.OneToOneField("Connection", on_delete=models.CASCADE, related_name="wallet")

    def __str__(self) -> str:  # pragma: no cover
        return f"WALLET: {self.pk=}. OWNER: {self.owner=}"

    @property
    def space(self) -> "Space":
        """Return the space that owns the wallet."""
        return self.owner.space

    @property
    def exchange_account(self) -> "ExchangeAccount":
        """Return the exchange account that contains the wallet."""
        return self.space.exchange_account.find()



import json
from typing import TYPE_CHECKING
from unittest import skipIf

from django_napse.core.models import BinanceAccount, Exchange, ExchangeAccount
from django_napse.core.settings import napse_settings
from django_napse.utils.errors import ExchangeAccountError
from django_napse.utils.model_test_case import ModelTestCase

if TYPE_CHECKING:
    from pathlib import Path

"""
python tests/test_app/manage.py test tests.django_tests.db.accounts.test_exchange -v2 --keepdb --parallel
"""


class ExchangeTestCase(ModelTestCase):
    model = Exchange
    skip_exchange_validation = True

    def simple_create(self):
        return self.model.objects.create(
            name="random exchange",
            description="random description",
        )

    def test_default_exchanges(self):
        base_configs = napse_settings.NAPSE_EXCHANGE_CONFIGS
        all_exchanges = Exchange.objects.all()

        self.assertEqual([exchange.name for exchange in all_exchanges], list(base_configs.keys()))

    def test_default_exchange_accounts(self):
        path: Path = napse_settings.NAPSE_SECRETS_FILE_PATH
        json_file = path.open("r")
        secrets = json.load(json_file)
        json_file.close()
        for exchange_id, exchange_secrets in secrets["Exchange Accounts"].items():
            exchange_name = exchange_secrets.pop("exchange")
            exchange = Exchange.objects.get(name=exchange_name)
            exchange_account = ExchangeAccount.objects.get(exchange=exchange, name=exchange_id)

            self.assertEqual(exchange_account.name, exchange_id)
            self.assertEqual(exchange_account.exchange.name, exchange_name)

            exchange_account_real_model = exchange_account.find()

            for key, value in exchange_secrets.items():
                self.assertEqual(getattr(exchange_account_real_model, key), value)


class BaseExchangeAccountTestCase(ModelTestCase):
    model = ExchangeAccount
    skip_exchange_validation = True

    def simple_create(self):
        exchange = Exchange.objects.create(
            name="random exchange",
            description="random description",
        )
        return self.model.objects.create(
            exchange=exchange,
            testing=True,
            name="random exchange account 1",
            description="random description",
        )


class ExchangeUtilsTestCase:
    @skipIf(napse_settings.NAPSE_IS_IN_PIPELINE, "No creadentials in pipelines")
    def test_ping(self):
        for exchange_account in self.model.objects.all():
            exchange_account.ping()

    def test_enough_accounts(self):
        self.assertTrue(self.model.objects.count() >= 1)

    def test_error_in_ping(self):
        with self.assertRaises(ExchangeAccountError.APIPermissionError):
            self.simple_create().ping()


class AccountUtilsBINANCETestCase(ExchangeUtilsTestCase, ModelTestCase):
    model = BinanceAccount
    exchange = "BINANCE"

    def simple_create(self):
        exchange = Exchange.objects.create(
            name="random exchange",
            description="random description",
        )
        return self.model.objects.create(
            exchange=exchange,
            testing=True,
            name="random exchange account 1",
            description="random description",
            private_key="random private key",
            public_key="random public key",
        )



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models import Currency
from django_napse.core.models.bots.controller import Controller


class CurrencySerializer(serializers.ModelSerializer):
    """Serialize a currency instance."""

    value = serializers.SerializerMethodField(read_only=True)

    class Meta:  # noqa: D106
        model = Currency
        fields: ClassVar[list[str]] = [
            "mbp",
            "ticker",
            "amount",
            "value",
        ]
        read_only_fields = fields

    def get_value(self, instance):
        return instance.amount * Controller.get_asset_price(
            exchange_account=instance.wallet.exchange_account,
            base=instance.ticker,
        )



from django.apps import AppConfig


class DjangoNapseSimulationsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_napse.simulations"
    label = "django_napse_simulations"



class ExchangeError:
    """Base class for exchange errors."""

    class UnavailableTicker(Exception):
        """Raised when a ticker is not available on the exchange."""


class ExchangeAccountError:
    """Base class for exchange account errors."""

    class CreateError(Exception):
        """Raised when a exchange account cannot be created."""

    class APIPermissionError(Exception):
        """Raised when a exchange account cannot interact with the exchange API."""



"""napse URL Configuration.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.2/topics/http/urls/

Examples:
    Function views
        1. Add an import:  from my_app import views
        2. Add a URL to urlpatterns:  path('', views.home, name='home')
    Class-based views
        1. Add an import:  from other_app.views import Home
        2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
    Including another URLconf
        1. Import the include() function: from django.urls import include, path
        2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))

"""

from django.conf.urls import include
from django.contrib import admin
from django.urls import path

from django_napse.api.api_urls import main_api_router

urlpatterns = [
    path("admin/", admin.site.urls),
    path("api/", include(main_api_router.urls), name="api"),
]



from django.db import models

from django_napse.core.models.transactions.managers import TransactionManager
from django_napse.utils.constants import TRANSACTION_TYPES


class Transaction(models.Model):
    """A Transaction is a transfer of value between two wallets."""

    from_wallet = models.ForeignKey("Wallet", on_delete=models.CASCADE, related_name="transactions_from")
    to_wallet = models.ForeignKey("Wallet", on_delete=models.CASCADE, related_name="transactions_to")
    amount = models.FloatField()
    ticker = models.CharField(max_length=10)
    transaction_type = models.CharField(max_length=20, default=TRANSACTION_TYPES.TRANSFER)
    created_at = models.DateTimeField(auto_now_add=True)

    objects = TransactionManager()

    def __str__(self) -> str:
        return f"TRANSACTION: {self.from_wallet.pk} -> {self.to_wallet.pk} ({self.amount=} - {self.ticker=})"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Transaction {self.pk=}\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.amount=}\n"
        string += f"{beacon}\t{self.ticker=}\n"
        string += f"{beacon}\t{self.transaction_type=}\n"
        string += f"{beacon}From Wallet:\n"
        wall_str = self.from_wallet.info(verbose=False, beacon=beacon + "\t")
        string += f"{wall_str}\n"
        string += f"{beacon}To Wallet:\n"
        wall_str = self.to_wallet.info(verbose=False, beacon=beacon + "\t")
        string += f"{wall_str}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string



from __future__ import annotations

from typing import ClassVar

from rest_framework import serializers

from django_napse.api.histories.serializers.datapoint import HistoryDataPointSerializer
from django_napse.core.models.histories.history import History


class HistorySerializer(serializers.ModelSerializer):
    """Serialize a wallet instance."""

    data_points = HistoryDataPointSerializer(many=True)

    class Meta:  # noqa: D106
        model = History
        fields: ClassVar[list[str]] = [
            "uuid",
            "data_points",
        ]



from django_napse.core.models import Credit, ExchangeAccount, Space, Transaction
from django_napse.utils.constants import TRANSACTION_TYPES
from django_napse.utils.errors import TransactionError
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.transactions.test_transactions -v2 --keepdb --parallel
"""


class TransactionTestCase:
    model = Transaction

    def setUp(self):
        self.from_wallet = self.space.wallet
        space = Space.objects.create(
            name="Test Space 2",
            exchange_account=self.exchange_account,
            description="This is a test space",
        )
        Credit.objects.create(wallet=self.from_wallet, amount=10, ticker="BTC")
        Credit.objects.create(wallet=space.wallet, amount=10, ticker="BTC")
        self.to_wallet = space.wallet

    def simple_create(self):
        return Transaction.objects.create(
            from_wallet=self.from_wallet,
            to_wallet=self.to_wallet,
            amount=1,
            ticker="BTC",
            transaction_type=TRANSACTION_TYPES.TRANSFER,
        )

    def test_error_same_space(self):
        exchange_account2 = ExchangeAccount.objects.create(
            exchange=self.exchange_account.exchange,
            testing=True,
            name="random exchange account 2",
            description="random description",
        )
        space2 = Space.objects.create(
            name="Test Space 2",
            exchange_account=exchange_account2,
            description="This is a test space",
        )
        wallet = space2.wallet
        with self.assertRaises(TransactionError.DifferentAccountError):
            Transaction.objects.create(
                from_wallet=self.from_wallet,
                to_wallet=wallet,
                amount=1,
                ticker="BTC",
                transaction_type=TRANSACTION_TYPES.TRANSFER,
            )

    def test_amount_zero(self):
        transaction = Transaction.objects.create(
            from_wallet=self.from_wallet,
            to_wallet=self.to_wallet,
            amount=0,
            ticker="BTC",
            transaction_type=TRANSACTION_TYPES.TRANSFER,
        )
        self.assertIsNone(transaction)

    def test_error_transaction_type(self):
        with self.assertRaises(TransactionError.InvalidTransactionError):
            Transaction.objects.create(
                from_wallet=self.from_wallet,
                to_wallet=self.to_wallet,
                amount=1,
                ticker="BTC",
                transaction_type="random transaction type",
            )


class TransactionBINANCETestCase(TransactionTestCase, ModelTestCase):
    exchange = "BINANCE"



from django.test import TestCase
from rest_framework.serializers import ValidationError

from django_napse.core.models import Exchange, ExchangeAccount
from django_napse.utils.serializers import MetaSerializer, Serializer
from django_napse.utils.serializers.fields import BoolField, DatetimeField, IntField, MethodField, StrField, UUIDField

"""
python tests/test_app/manage.py test tests.django_tests.test_serializer -v2 --keepdb --parallel
"""


class ExchangeSerializer(Serializer):
    Model = Exchange
    name = StrField(required=True)
    description = StrField()
    test_int_field = IntField()


class ExchangeAccountSerializer(Serializer):
    Model = ExchangeAccount
    uuid = UUIDField()
    exchange = ExchangeSerializer(required=True)
    name = StrField(required=True)
    testing = BoolField(required=True)
    description = StrField()
    default = BoolField()
    created_at = DatetimeField()
    tickers = MethodField()

    def get_tickers(self, instance):
        return instance.exchange.get_tickers()

    def validate_data(self, data):
        data["testing"] = True
        return super().validate_data(data)


class SerializerSerializationTestCase(TestCase):
    def setUp(self):
        self.exchange = Exchange.objects.create(name="Binance", description="Binance exchange")
        self.exchange.test_int_field = 1  # Just for testing the int field
        self.exchange_account = ExchangeAccount.objects.create(
            exchange=self.exchange,
            name="Binance Account",
            testing=True,
            description="Binance account description",
            default=True,
        )

    def test_exchange_serialization(self):
        self.exchange.test_int_field = 1  # Just for testing the int field
        exchange_serializer = ExchangeSerializer(instance=self.exchange)
        data = exchange_serializer.data
        self.assertEqual(
            data,
            {
                "name": "Binance",
                "description": "Binance exchange",
                "test_int_field": 1,
            },
        )

    def test_exchange_account_serialization(self):
        exchange_account_serializer = ExchangeAccountSerializer(self.exchange_account)
        data = exchange_account_serializer.data
        self.assertEqual(
            data,
            {
                "uuid": str(self.exchange_account.uuid),
                "exchange": {
                    "name": "Binance",
                    "description": "Binance exchange",
                    "test_int_field": 1,
                },
                "name": "Binance Account",
                "testing": True,
                "description": "Binance account description",
                "default": True,
                "created_at": self.exchange_account.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                "tickers": [],
            },
        )

    def test_many_serialization(self):
        instances = [self.exchange for _ in range(10)]
        exchange_serializer = ExchangeSerializer(many=True, instance=instances)
        data = exchange_serializer.data
        self.assertEqual(len(data), 10)

    def test_source_serialization(self):
        exchange_account_serializer = ExchangeAccountSerializer
        exchange_account_serializer.fields_map["exchange"] = StrField(source="exchange.name")
        exchange_account_serializer.compiled_fields = MetaSerializer._compile_fields(  # noqa: SLF001
            fields=exchange_account_serializer.fields_map,
            serializer_cls=exchange_account_serializer,
        )

        exchange_account_serializer = exchange_account_serializer(self.exchange_account)
        data = exchange_account_serializer.data
        self.assertEqual(data.get("exchange"), self.exchange.name)

        # reset
        exchange_account_serializer = ExchangeAccountSerializer
        exchange_account_serializer.fields_map["exchange"] = ExchangeSerializer()
        exchange_account_serializer.compiled_fields = MetaSerializer._compile_fields(  # noqa: SLF001
            fields=exchange_account_serializer.fields_map,
            serializer_cls=exchange_account_serializer,
        )

    def test_many_source_serialization(self):
        exchange_account_serializer = ExchangeAccountSerializer
        exchange_account_serializer.fields_map["exchange"] = StrField(source="exchange.name")
        exchange_account_serializer.compiled_fields = MetaSerializer._compile_fields(  # noqa: SLF001
            fields=exchange_account_serializer.fields_map,
            serializer_cls=exchange_account_serializer,
        )
        exchange_account_serializer = exchange_account_serializer(instance=[self.exchange_account for _ in range(10)], many=True)
        data = exchange_account_serializer.data
        self.assertEqual(len(data), 10)

        # reset
        exchange_account_serializer = ExchangeAccountSerializer
        exchange_account_serializer.fields_map["exchange"] = ExchangeSerializer()
        exchange_account_serializer.compiled_fields = MetaSerializer._compile_fields(  # noqa: SLF001
            fields=exchange_account_serializer.fields_map,
            serializer_cls=exchange_account_serializer,
        )


class SerializerValidationTestCase(TestCase):
    def setUp(self):
        self.exchange = Exchange.objects.create(name="Binance", description="Binance exchange")
        self.exchange.test_int_field = 1  # Just for testing the int field
        self.exchange_account = ExchangeAccount.objects.create(
            exchange=self.exchange,
            name="Binance Account",
            testing=True,
            description="Binance account description",
            default=True,
        )

    def test_exchange_validation_and_creation(self):
        data = {"name": "TEST", "description": "Test exchange"}
        exchange_serializer = ExchangeSerializer(data=data)
        instance = exchange_serializer.create()
        self.assertEqual(isinstance(instance, Exchange), True)
        self.assertEqual(exchange_serializer.data, data)

    def test_exchange_account_validation_and_creation(self):
        data = {
            "name": "TEST exchange account",
            "description": "Test exchange",
            "exchange": self.exchange.id,
            "testing": True,
        }
        exchange_serializer = ExchangeAccountSerializer(data=data)
        instance = exchange_serializer.create()
        self.assertEqual(isinstance(instance, ExchangeAccount), True)
        self.assertEqual(
            isinstance(
                exchange_serializer.validated_data.get("exchange"),
                Exchange,
            ),
            True,
        )

    def test_exchange_validation_and_update(self):
        data = {"name": "TEST", "description": "new description"}
        exchange_serializer = ExchangeSerializer(data=data)
        instance = exchange_serializer.update(instance=self.exchange)
        self.assertEqual(isinstance(instance, Exchange), True)
        self.assertEqual(instance.description, "new description")

    def test_validation_with_missing_required_field(self):
        with self.assertRaises(ValidationError):
            exchange_serializer = ExchangeSerializer(data={"description": "Test exchange"})  # noqa: F841

    def test_validation_with_invalid_field(self):
        data = {
            "name": "TEST exchange account",
            "description": "Test exchange",
            "exchange": {"name": 12},
            "testing": True,
        }
        with self.assertRaises(ValidationError):
            ExchangeAccountSerializer(data=data)

        data = {
            "name": "TEST",
            "description": "Test exchange",
            "test_int_field": "whatever",
        }
        with self.assertRaises(ValidationError):
            ExchangeSerializer(data=data)

    def test_validation_with_invalid_data(self):
        data = 1
        with self.assertRaises(ValidationError):
            ExchangeSerializer(data=data)

    def test_validation_overwrite(self):
        data = {
            "name": "TEST exchange account",
            "description": "Test exchange",
            "exchange": self.exchange.id,
            "testing": False,
        }
        exchange_serializer = ExchangeAccountSerializer(data=data)
        instance = exchange_serializer.create()
        self.assertEqual(isinstance(instance, ExchangeAccount), True)
        self.assertEqual(
            exchange_serializer.validated_data,
            {
                "name": "TEST exchange account",
                "description": "Test exchange",
                "exchange": self.exchange,
                "testing": True,
            },
        )

    def test_get_validated_data_before_validation(self):
        exchange_serializer = ExchangeSerializer()
        with self.assertRaises(ValueError):
            exchange_serializer.validated_data  # noqa: B018

    def test_model_action_before_validation(self):
        exchange_serializer = ExchangeSerializer()
        with self.assertRaises(ValueError):
            exchange_serializer.create()

    def test_model_action_without_model(self):
        exchange_serializer = ExchangeSerializer()
        exchange_serializer.Model = None
        with self.assertRaises(ValueError):
            exchange_serializer.create({})

    def _test_perf(self):
        instances = [self.exchange_account for _ in range(1_000_000)]
        from time import time

        start = time()
        serializer = ExchangeAccountSerializer(instance=instances, many=True)
        print(f"setup: {time() - start:0.3f}s")
        data = serializer.data
        print(f"Total time{time() - start:0.6f}s")
        self.assertEqual(len(data), 1_000_000)



from typing import ClassVar

from rest_framework import serializers

from django_napse.api.spaces.serializers.space_serializers import SpaceSerializer
from django_napse.core.models import ExchangeAccount


class ExchangeAccountDetailSerializer(serializers.ModelSerializer):
    """Serializer for ExchangeAccount detail."""

    exchange = serializers.CharField(source="exchange.name")
    spaces = SpaceSerializer(many=True, read_only=True)

    class Meta:  # noqa: D106
        model = ExchangeAccount
        fields: ClassVar[list[str]] = [
            "uuid",
            "exchange",
            "name",
            "testing",
            "description",
            "created_at",
            "spaces",
        ]
        read_only_fields: ClassVar[list[str]] = fields



from django.apps import apps
from django.db import models


class SimulationDataPointManager(models.Manager):
    def create(self, simulation, date, value: float, action: str, amount: float, ticker: str, extra_info=None):
        SimulationDataPointExtraInfo = apps.get_model("django_napse_simulations", "SimulationDataPointExtraInfo")

        extra_info = extra_info or {}

        data_point = self.model(
            simulation=simulation,
            date=date,
            value=value,
            action=action,
            amount=amount,
            ticker=ticker,
        )
        data_point.save()
        for key, value in extra_info.items():
            print(key, value)
            SimulationDataPointExtraInfo.objects.create(
                data_point=data_point,
                key=key,
                value=str(value),
                target_type=type(value).__name__,
            )
        return data_point



class TransactionError:
    """Base class for transaction errors."""

    class DifferentAccountError(Exception):
        """Raised when the from and to wallets are on different spaces."""

    class InvalidTransactionError(Exception):
        """Raised when the transaction is invalid."""



from __future__ import annotations

from django.db import models

from django_napse.core.models.bots.managers.strategy import StrategyManager
from django_napse.utils.findable_class import FindableClass


class Strategy(models.Model, FindableClass):
    """Define the bot's buying and selling logic."""

    config = models.OneToOneField("BotConfig", on_delete=models.CASCADE, related_name="strategy")
    architecture = models.OneToOneField("Architecture", on_delete=models.CASCADE, related_name="strategy")

    objects = StrategyManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"STRATEGY {self.pk}"

    def info(self, verbose: bool = True, beacon: str = "") -> str:  # noqa: FBT001, FBT002
        """Return info on the Strategy."""
        string = ""
        string += f"{beacon}Strategy {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.config=}\n"
        string += f"{beacon}\t{self.architecture=}\n"

        if verbose:
            print(string)
        return string

    def give_order(self, data: dict) -> list[dict]:  # noqa
        if self.__class__ == Strategy:
            error_msg = "give_order not implemented for the Strategy base class, please implement it in a subclass."
        else:
            error_msg = f"give_order not implemented for the Strategy base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    @classmethod
    def plugin_classes(cls) -> list:
        """Return the list of all strategy's plugins."""
        return []

    @classmethod
    def achitecture_class(cls):  # noqa
        if cls == Strategy:
            error_msg = "achitecture_class not implemented for the Strategy base class, please implement it in a subclass."
        else:
            error_msg = f"achitecture_class not implemented for the Strategy base class, please implement it in the {cls} class."
        raise NotImplementedError(error_msg)

    @classmethod
    def config_class(cls):  # noqa
        if cls == Strategy:
            error_msg = "config_class not implemented for the Strategy base class, please implement it in a subclass."
        else:
            error_msg = f"config_class not implemented for the Strategy base class, please implement it in the {cls} class."
        raise NotImplementedError(error_msg)

    @classmethod
    def architecture_class(cls) -> "Architecture":  # noqa: F821
        """Return the class of the associated architecture."""
        return cls._meta.get_field("architecture").related_model

    @property
    def variables(self) -> dict[str, any]:
        """Return variables of a strategy."""
        self = self.find(self.pk)
        variables = {}
        for variable in self._meta.get_fields():
            if variable.name.startswith("variable_"):
                variables[variable.name[8:]] = getattr(self, variable.name)
        return variables

    def connect(self, connection):  # noqa
        return

    def copy(self) -> Strategy:
        """Return a copy of the Strategy."""
        return self.find().__class__.objects.create(
            config=self.config.find().duplicate_immutable(),
            architecture=self.architecture.find().copy(),
        )



from typing import TYPE_CHECKING, Optional

from django.apps import apps
from django.db import models

if TYPE_CHECKING:
    from django_napse.core.models.histories.history import History, HistoryDataPoint


class HistoryDataPointManager(models.Manager):
    """The manager for the HistoryDataPoint model."""

    def create(
        self,
        history: "History",
        points: Optional[dict] = None,
    ) -> "HistoryDataPoint":
        """Create a new data point for the history."""
        HistoryDataPointField = apps.get_model("django_napse_core", "HistoryDataPointField")
        points = points or {}
        data_point = self.model(history=history)
        data_point.save()
        for key, value in points.items():
            HistoryDataPointField.objects.create(history_data_point=data_point, key=key, value=value)
        return data_point



from django_napse.core.models.bots.plugin import Plugin
from django_napse.core.models.connections.connection import ConnectionSpecificArgs
from django_napse.core.models.wallets.currency import CurrencyPydantic
from django_napse.utils.constants import PLUGIN_CATEGORIES, SIDES


class SBVPlugin(Plugin):
    @classmethod
    def plugin_category(cls):
        return PLUGIN_CATEGORIES.POST_ORDER

    def apply__no_db(self, data: dict) -> dict:
        order = data["order"]
        current_base_amout = (
            data["connection_data"][data["connection"]]["wallet"]
            .currencies.get(order["controller"].base, CurrencyPydantic(ticker=order["controller"].base, amount=0, mbp=0))
            .amount
        )
        current_quote_amout = (
            data["connection_data"][data["connection"]]["wallet"]
            .currencies.get(order["controller"].quote, CurrencyPydantic(ticker=order["controller"].quote, amount=0, mbp=0))
            .amount
        )
        if data["connection_data"][data["connection"]]["connection_specific_args"]["sbv"].get_value() is None or order["side"] == SIDES.SELL:
            order["ConnectionModifications"] += [
                {
                    "key": "sbv",
                    "value": current_quote_amout + current_base_amout * order["price"],
                    "target_type": "float",
                    "ignore_failed_order": False,
                    "connection_specific_arg": data["connection_data"][data["connection"]]["connection_specific_args"]["sbv"],
                },
            ]
        return data

    def _connect(self, connection):
        ConnectionSpecificArgs.objects.create(connection=connection, key="sbv", value="None", target_type="float")



from typing import TYPE_CHECKING

from django.db import models

if TYPE_CHECKING:
    from django_napse.core.models.histories.history import HistoryDataPoint, HistoryDataPointField


class HistoryDataPointFieldManager(models.Manager):
    """The manager for the HistoryDataPointField model."""

    def create(
        self,
        history_data_point: "HistoryDataPoint",
        key: str,
        value: str,
    ) -> "HistoryDataPointField":
        """Create a new data point field for a data point."""
        data_point_field = self.model(history_data_point=history_data_point, key=key, value=str(value), target_type=type(value).__name__)
        data_point_field.save()
        return data_point_field



from django_napse.core.models.histories.history import History
from django_napse.core.tasks.base_task import BaseTask


class HistoryUpdateTask(BaseTask):
    """Task to update all histories."""

    name = "history_update"
    interval_time = 3600
    time_limit = 60
    soft_time_limit = 60

    def run(self) -> None:
        """Run a task to update all controllers."""
        if not self.avoid_overlap(verbose=False):
            return
        self.info("Running HistoryUpdateTask")
        for history in History.objects.all():
            history.find().generate_data_point()


HistoryUpdateTask().delete_task()
HistoryUpdateTask().register_task()



import uuid
from datetime import datetime, timedelta

from django.db import models
from django.utils.timezone import get_default_timezone

from django_napse.core.models.bots.bot import Bot
from django_napse.core.models.connections.connection import Connection
from django_napse.core.models.fleets.managers import FleetManager
from django_napse.core.models.orders.order import Order
from django_napse.utils.errors import BotError


class Fleet(models.Model):
    uuid = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        unique=True,
    )
    name = models.CharField(
        max_length=100,
        default="Fleet",
    )
    exchange_account = models.ForeignKey(
        "ExchangeAccount",
        on_delete=models.CASCADE,
    )
    running = models.BooleanField(default=False)
    setup_finished = models.BooleanField(default=False)
    created_at = models.DateTimeField(
        auto_now_add=True,
        blank=True,
    )

    objects = FleetManager()

    def __str__(self):
        return f"FLEET: {self.pk=}, name={self.name}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Fleet {self.pk}:\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.name=}\n"
        string += f"{beacon}\t{self.exchange_account=}\n"
        string += f"{beacon}\t{self.running=}\n"
        string += f"{beacon}\t{self.setup_finished=}\n"

        string += f"{beacon}Clusters:\n"
        new_beacon = beacon + "\t"
        for cluster in self.clusters.all():
            string += f"{beacon}{cluster.info(verbose=False, beacon=new_beacon)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self):
        return self.exchange_account.testing

    @property
    def bots(self):
        return Bot.objects.filter(link__cluster__fleet=self)

    @property
    def value(self) -> float:
        """Sum value of all bots in fleet."""
        connections = Connection.objects.filter(bot__in=self.bots)
        return sum([connection.wallet.value_market() for connection in connections])

    def space_frame_value(self, space) -> float:
        """Sum value of all bots connected to the space."""
        # TODO: remove property to values and add the following lines to the new `value()` method
        fleet_connections = Connection.objects.filter(bot__in=self.bots)
        space_connections = space.wallet.connections.all()
        commun_connections = space_connections.intersection(fleet_connections)
        return sum([connection.wallet.value_market() for connection in commun_connections])

    def bot_clusters(self):
        bot_clusters = []
        for cluster in self.clusters.all():
            bot_clusters.append(Bot.objects.filter(link__cluster=cluster))
        return bot_clusters

    def connect_to_space(self, space):
        ...

    def invest(self, space, amount, ticker):
        connections = []
        for cluster in self.clusters.all():
            connections += cluster.invest(space, amount * cluster.share, ticker)
        return connections

    def bot_count(self, space=None) -> int:
        """Count number of bots in fleet, depends on space frame."""
        query_bot = self.bots.all()
        if space is None:
            return len(query_bot)
        result = []
        for bot in query_bot:
            try:
                bot_space = bot.space
            except BotError.NoSpace:
                continue
            if bot_space == self.space:
                result.append(bot)
        return len(result)

    def get_stats(self) -> dict[str, str | int | float]:
        """Return Fleet's stats."""
        order_count = Order.objects.filter(  # noqa: F841
            connection__bot__in=self.bots,
            created_at__gt=datetime.now(tz=get_default_timezone()) - timedelta(days=30),
        ).count()
        return {
            "value": self.value,
            "bot_count": self.bot_count(),
            "delta_30": 0,  # TODO: Need history
        }

    def delete(self) -> None:
        """Delete clusters & relative (template) bots."""
        self.bots.delete()
        self.clusters.all().delete()
        super().delete()



import uuid
from datetime import datetime

from pytz import UTC

from django_napse.core.models import Bot, Controller, EmptyBotConfig, EmptyStrategy, SinglePairArchitecture
from django_napse.simulations.models import Simulation
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.simulations.test_simulation -v2 --keepdb --parallel
"""


class SimulationTestCase:
    model = Simulation

    def simple_create(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        architecture = SinglePairArchitecture.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1m",
                ),
            },
        )
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        return Simulation.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 1, 3, tzinfo=UTC),
            simulation_reference=uuid.uuid4(),
            data={},
        )

    def test_with_data(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        architecture = SinglePairArchitecture.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1m",
                ),
            },
        )
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        simulation = Simulation.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 1, 3, tzinfo=UTC),
            simulation_reference=uuid.uuid4(),
            data={
                "dates": [datetime.now(tz=UTC)],
                "values": [1],
                "actions": ["BUY"],
                "amounts": [1],
                "tickers": ["USDT"],
                "mbp": [10],
            },
        )
        simulation.info(verbose=False)

    def test_with_more_data(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        architecture = SinglePairArchitecture.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1m",
                ),
            },
        )
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        simulation = Simulation.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2021, 1, 1, tzinfo=UTC),
            end_date=datetime(2021, 1, 3, tzinfo=UTC),
            simulation_reference=uuid.uuid4(),
            data={
                "dates": [
                    datetime(2021, 1, 1, tzinfo=UTC),
                    datetime(2021, 1, 2, tzinfo=UTC),
                    datetime(2021, 1, 3, tzinfo=UTC),
                    datetime(2021, 1, 4, tzinfo=UTC),
                    datetime(2021, 1, 5, tzinfo=UTC),
                    datetime(2021, 1, 6, tzinfo=UTC),
                    datetime(2021, 1, 7, tzinfo=UTC),
                    datetime(2021, 1, 8, tzinfo=UTC),
                    datetime(2021, 1, 9, tzinfo=UTC),
                    datetime(2021, 1, 10, tzinfo=UTC),
                ],
                "values": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                "actions": ["BUY", "SELL", "SELL", "BUY", "KEEP", "KEEP", "BUY", "SELL", "BUY", "SELL"],
                "amounts": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                "tickers": ["USDT", "USDT", "USDT", "USDT", "USDT", "USDT", "USDT", "USDT", "USDT", "USDT"],
                "mbp": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
            },
        )
        simulation.info(verbose=False)


class SimulationBINANCETestCase(SimulationTestCase, ModelTestCase):
    exchange = "BINANCE"



import uuid

from django.db import models

from django_napse.core.models.bots.managers.bot_config import BotConfigManager
from django_napse.core.models.bots.strategy import Strategy
from django_napse.utils.findable_class import FindableClass


class BotConfig(models.Model, FindableClass):
    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    space = models.ForeignKey("Space", on_delete=models.CASCADE, related_name="bot_configs")
    immutable = models.BooleanField(default=False)

    objects = BotConfigManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"BOT CONFIG: {self.pk}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}BotConfig: ({self.pk=})\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.space=}\n"
        string += f"{beacon}Settings:\n"
        for setting, value in self.settings.items():
            string += f"{beacon}\t" + setting + f"={value}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def settings(self):
        settings = {}
        for setting in self._meta.get_fields():
            if setting.name.startswith("setting_"):
                settings[setting.name[8:]] = getattr(self, setting.name)
        return settings

    @property
    def strategy(self):
        return Strategy.objects.get(config=self).find()

    def duplicate_immutable(self):
        return self.__class__.objects.create(
            space=self.space,
            immutable=True,
            settings=self.settings,
        )

    def duplicate_other_space(self, space):
        if space == self.space:
            error_msg = "The space of the new bot config must be different from the space of the original bot config."
            raise ValueError(error_msg)
        return self.__class__.objects.create(
            space=space,
            immutable=True,
            settings=self.settings,
        )



from django_napse.core.models import Credit
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.transactions.test_credit -v2 --keepdb --parallel
"""


class CreditTestCase:
    model = Credit

    def simple_create(self):
        return Credit.objects.create(
            wallet=self.space.wallet,
            amount=1,
            ticker="BTC",
        )

    def test_empty_credit(self):
        credit = Credit.objects.create(
            wallet=self.space.wallet,
            amount=0,
            ticker="BTC",
        )
        self.assertIsNone(credit)

    def test_correct_amount(self):
        self.simple_create()
        self.assertEqual(self.space.wallet.get_amount("BTC"), 1)
        self.simple_create()
        self.assertEqual(self.space.wallet.get_amount("BTC"), 2)


class CreditBINANCETestCase(CreditTestCase, ModelTestCase):
    exchange = "BINANCE"



import uuid
from json import loads
from typing import ClassVar

from rest_framework import serializers

from django_napse.api.fleets.serializers import FleetSerializer
from django_napse.api.wallets.serializers.wallet_serializers import WalletSerializer
from django_napse.core.models import ExchangeAccount, Space, SpaceHistory


class SpaceSerializer(serializers.ModelSerializer):
    """To serialize & create a space instance."""

    exchange_account = serializers.CharField(source="exchange_account.uuid")
    delta = serializers.SerializerMethodField(read_only=True)

    class Meta:  # noqa: D106
        model = Space
        fields: ClassVar[list[str]] = [
            "name",
            "description",
            "exchange_account",
            # read-only
            "uuid",
            "testing",
            "value",
            "delta",
        ]
        read_only_fields: ClassVar[list[str]] = [
            "uuid",
            "testing",
            "value",
            "delta",
        ]

    def get_delta(self, instance: Space) -> float:
        """Delta on the last 30 days."""
        try:
            history = SpaceHistory.objects.get(owner=instance)
        except SpaceHistory.DoesNotExist:
            return 0
        return history.get_delta()

    def create(self, validated_data: dict[str, any]) -> Space:
        """Create a space if data are validated."""
        try:
            uuid.UUID(str(validated_data["exchange_account"]["uuid"]))
        except ValueError:
            error_msg: str = "Invalid UUID"
            raise serializers.ValidationError(error_msg) from None
        try:
            exchange_account = ExchangeAccount.objects.get(uuid=validated_data["exchange_account"]["uuid"])
        except ExchangeAccount.DoesNotExist:
            error_msg: str = "Exchange Account does not exist"
            raise serializers.ValidationError(error_msg) from None

        return Space.objects.create(
            name=validated_data["name"],
            description=validated_data["description"],
            exchange_account=exchange_account,
        )


class SpaceDetailSerializer(serializers.ModelSerializer):
    """Deep serialization of a space instance."""

    fleets = FleetSerializer(many=True, read_only=True)
    exchange_account = serializers.CharField(source="exchange_account.uuid", read_only=True)
    statistics = serializers.SerializerMethodField(read_only=True)
    wallet = WalletSerializer(read_only=True)
    history = serializers.SerializerMethodField(read_only=True)

    class Meta:  # noqa: D106
        model = Space
        fields: ClassVar[list[str]] = [
            "name",
            "description",
            # read-only
            "uuid",
            "testing",
            "exchange_account",
            "created_at",
            "statistics",
            "wallet",
            "history",
            "fleets",
        ]
        read_only_fields: ClassVar[list[str]] = fields

    def get_statistics(self, instance: Space) -> dict:
        """Return statistics of the given space."""
        return instance.get_stats()

    def get_history(self, instance: Space) -> list:
        """Return history of the given space."""
        try:
            history = SpaceHistory.objects.get(owner=instance)
        except SpaceHistory.DoesNotExist:
            return []

        return loads(history.to_dataframe().to_json(orient="records"))


class SpaceMoneyFlowSerializer(serializers.Serializer):
    """Serialization for money actions on the given space."""

    amount = serializers.FloatField(write_only=True, required=True)
    ticker = serializers.CharField(write_only=True, required=True)

    def __init__(
        self,
        side: str,
        *args: list[str],
        **kwargs: dict[str, any],
    ) -> None:
        """Definie side & initilize the serializer."""
        self.side = side
        super().__init__(*args, **kwargs)

    def _invest_validate(self, attrs: dict[str, any]) -> dict[str, any]:
        if not self.instance.testing:
            error_msg: str = "Not implemented yet."
            raise NotImplementedError(error_msg)

        # Test invest
        return attrs

    def _withdraw_validate(self, attrs: dict[str, any]) -> dict[str, any]:
        if self.instance.testing:
            error_msg: str = "Not implemented yet."
            raise NotImplementedError(error_msg)

        # Test withdraw
        return attrs

    def validate(self, attrs: dict[str, any]) -> dict[str, any]:
        """Make validation."""
        if attrs.get("amount") <= 0:
            error_msg: str = "Invalid amount."
            raise serializers.ValidationError(error_msg)

        if attrs.get("ticker") not in self.instance.exchange_account.get_tickers():
            error_msg: str = f"{attrs['ticker']} is not available on {self.instance.exchange_account.exchange.name} exchange."
            raise serializers.ValidationError(error_msg)

        match self.side.upper():
            case "INVEST":
                return self._invest_validate(attrs)
            case "WITHDRAW":
                return self._withdraw_validate(attrs)
            case _:
                error_msg: str = "Invalid side."
                raise ValueError(error_msg)

    def save(self) -> None:
        """Make validated action on the space."""
        if self.side.upper() == "INVEST":
            self.instance.invest(
                self.validated_data.get("amount"),
                self.validated_data.get("ticker"),
            )
        else:
            self.instance.withdraw(self.amount, self.ticker)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework_api_key.permissions import HasAPIKey

from django_napse.api.custom_permissions import HasFullAccessPermission, HasMasterKey, HasReadPermission
from django_napse.api.custom_viewset import CustomViewSet
from django_napse.api.spaces.serializers import SpaceDetailSerializer, SpaceMoneyFlowSerializer, SpaceSerializer, serializers
from django_napse.core.models import Space
from django_napse.utils.constants import EXCHANGE_TICKERS
from django_napse.utils.errors import SpaceError

if TYPE_CHECKING:
    from uuid import UUID

    from django.db.models import QuerySet
    from rest_framework.request import Request


class SpaceView(CustomViewSet):
    """Endpoints for Spaces."""

    permission_classes: ClassVar[list] = [HasFullAccessPermission]
    serializer_class = SpaceSerializer

    def get_object(self) -> Space:
        """Return the space instance for detail endpoint."""
        return self.get_queryset().get(uuid=self.kwargs["pk"])

    def get_queryset(self) -> QuerySet[Space]:
        """Return all spaces instances."""
        return Space.objects.all()

    def get_serializer_class(self) -> serializers.Serializer:
        """Return the serializer class for each action."""
        actions: dict = {
            "list": SpaceSerializer,
            "retrieve": SpaceDetailSerializer,
            "create": SpaceSerializer,
            "update": SpaceSerializer,
            "partial_update": SpaceSerializer,
            "invest": SpaceMoneyFlowSerializer,
            "withdraw": SpaceMoneyFlowSerializer,
        }
        result = actions.get(self.action, None)
        return result if result else super().get_serializer_class()

    def get_permissions(self) -> list:
        """Return permissions for each action."""
        match self.action:
            case "retrieve":
                return [HasReadPermission()]
            case "list":
                return [HasAPIKey()]
            case "create":
                return [HasMasterKey()]

            case _:
                return super().get_permissions()

    def list(self, request: Request) -> None:
        """Return all available spaces (depending of the keys permissions)."""
        api_key = self.get_api_key(request)
        spaces = Space.objects.all() if api_key.is_master_key else Space.objects.filter(permissions__in=api_key.permissions.all())
        serializer = self.get_serializer(spaces, many=True)
        data = serializer.data
        return Response(data, status=status.HTTP_200_OK)

    def retrieve(
        self,
        request: Request,  # noqa: ARG002
        pk: int | str | UUID | None = None,  # noqa: ARG002
    ) -> None:
        """Return the detail info of the given space."""
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        data = serializer.data
        return Response(data, status=status.HTTP_200_OK)

    def create(self, request: Request) -> None:
        """Create a space."""
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        space = serializer.save()
        serialized_space = self.get_serializer(space)
        return Response(serialized_space.data, status=status.HTTP_201_CREATED)

    def update(self, request: Request, **kwargs: dict[str, any]) -> None:
        """Update the given space."""
        partial = kwargs.pop("partial", False)
        instance = self.get_object()

        serializer = self.get_serializer(instance=instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(status=status.HTTP_200_OK)

    def partial_update(self, request: Request, **kwargs: dict[str, any]) -> None:
        """Update the given space."""
        return self.update(request, partial=True, **kwargs)

    def delete(
        self,
        request: Request,  # noqa: ARG002
        pk: int | str | UUID | None = None,  # noqa: ARG002
    ) -> None:
        """Delete the given space."""
        instance = self.get_object()
        try:
            instance.delete()
        except SpaceError.DeleteError:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=True, methods=["GET", "POST"])
    def invest(
        self,
        request: Request,
        pk: int | str | UUID | None = None,  # noqa: ARG002
    ) -> None:
        """Endpoint to invest on space.

        GET: Return all {ticker: amount} which can be invested in the space.
        POST: Invest in the space.
        """
        space: Space = self.get_object()
        if not space.testing:
            return Response(status=status.HTTP_501_NOT_IMPLEMENTED)

        exchange_name: str = space.exchange_account.exchange.name

        match request.method:
            case "GET":
                possible_investments = [{"ticker": ticker, "amount": 1_000_000} for ticker in EXCHANGE_TICKERS.get(exchange_name)]
                return Response(possible_investments, status=status.HTTP_200_OK)

            case "POST":
                space = self.get_object()
                if not space.testing:
                    error_msg: str = "Investing in real is not allowed yet."
                    return Response(error_msg, status=status.HTTP_403_FORBIDDEN)
                serializer = self.get_serializer(data=request.data, instance=space, side="INVEST")
                serializer.is_valid(raise_exception=True)
                serializer.save()

            case _:
                return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)

        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["GET", "POST"])
    def withdraw(
        self,
        request: Request,
        pk: int | str | UUID | None = None,  # noqa: ARG002
    ) -> None:
        """Endpoint to withdraw on space.

        GET: Return all {ticker: amount} which can be withdrawn in the space.
        POST: Withdraw from the space.
        """
        space: Space = self.get_object()
        if not space.testing:
            return Response(status=status.HTTP_501_NOT_IMPLEMENTED)

        match request.method:
            case "GET":
                return Response(status=status.HTTP_200_OK)
            case "POST":
                return Response(status=status.HTTP_501_NOT_IMPLEMENTED)
            case _:
                return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)



from __future__ import annotations

from typing import ClassVar

from django.db.models import Q
from rest_framework import serializers

from django_napse.api.histories.serializers.history import HistorySerializer
from django_napse.api.transactions.serializers import CreditSerializer, DebitSerializer, TransactionSerializer
from django_napse.api.wallets.serializers.currency_serializer import CurrencySerializer
from django_napse.core.models import Transaction, Wallet


class WalletSerializer(serializers.ModelSerializer):
    """Serialize a wallet instance."""

    currencies = CurrencySerializer(many=True, read_only=True)
    value = serializers.SerializerMethodField(read_only=True)
    operations = serializers.SerializerMethodField(read_only=True)
    history = HistorySerializer(many=False, read_only=True)

    class Meta:  # noqa: D106
        model = Wallet
        fields: ClassVar[list[str]] = [
            "title",
            "value",
            "created_at",
            "currencies",
            "operations",
            "history",
        ]
        read_only_fields: ClassVar[list[str]] = [
            "value",
            "created_at",
            "currencies",
            "operations",
        ]

    def get_value(self, instance) -> float:
        return instance.value_market()

    def get_operations(self, instance) -> dict:
        transactions = Transaction.objects.filter(Q(from_wallet=instance) | Q(to_wallet=instance)).order_by("created_at")
        transactions_data = TransactionSerializer(transactions, many=True).data
        credits_data = CreditSerializer(instance.credits.all().order_by("created_at"), many=True).data
        debits_data = DebitSerializer(instance.debits.all().order_by("created_at"), many=True).data
        return credits_data + debits_data + transactions_data



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models import Credit


class CreditSerializer(serializers.ModelSerializer):
    """Serializer for Credit instance."""

    operation_type = serializers.CharField(default="CREDIT")

    class Meta:  # noqa: D106
        model = Credit
        fields: ClassVar[list[str]] = [
            "amount",
            "ticker",
            "operation_type",
            "created_at",
        ]
        read_only_fields = fields



from django_napse.core.models import Controller, SinglePairArchitecture
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_architecture -v2 --keepdb --parallel
"""


class ArchitectureTestCase:
    model = SinglePairArchitecture

    def simple_create(self):
        return self.model.objects.create(
            constants={
                "controller": Controller.get(
                    exchange_account=self.exchange_account,
                    base="BTC",
                    quote="USDT",
                    interval="1m",
                ),
            },
        )


class SinglePairArchitectureBINANCETestCase(ArchitectureTestCase, ModelTestCase):
    exchange = "BINANCE"



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models import ExchangeAccount


class ExchangeAccountSerializer(serializers.ModelSerializer):
    """Serializer for ExchangeAccount."""

    exchange = serializers.CharField(source="exchange.name")

    class Meta:  # noqa: D106
        model = ExchangeAccount
        fields: ClassVar[list[str]] = [
            "uuid",
            "exchange",
            "name",
            "description",
            "testing",
        ]
        read_only_fields: ClassVar[list[str]] = [
            "uuid",
            "exchange",
            "testing",
        ]



from datetime import datetime, timezone
from typing import Optional

import celery
from celery.utils.log import get_task_logger
from django.db import IntegrityError
from django.db.utils import ProgrammingError
from django_celery_beat.models import IntervalSchedule, PeriodicTask

from django_napse.core.celery_app import celery_app, strategy_log_free


class BaseTask(celery.Task):
    """Base class for all Celery tasks."""

    name = "base_task"
    Strategy = strategy_log_free
    logger = get_task_logger(name)
    interval_time = 5  # Impossible to make dynamic modification because of celery

    def __init__(self) -> None:
        super().__init__()
        self.logger.setLevel("INFO")

    def run(self) -> None:
        """Function called when running the task."""

    def info(self, msg: str) -> None:
        """Log a message."""
        info = f"[{self.name} @{datetime.now(tz=timezone.utc)}] : {msg}"
        self.logger.info(info)

    def error(self, msg: str) -> None:
        """Log an error."""
        error = f"[{self.name} @{datetime.now(tz=timezone.utc)}] : {msg}"
        self.logger.error(error)

    def warning(self, msg: str) -> None:
        """Log a warning."""
        warning = f"[{self.name} @{datetime.now(tz=timezone.utc)}] : {msg}"
        self.logger.warning(warning)

    def create_task(self) -> None:
        """Build task feed_bots_with_candles.

        Raises:
            ValidationError: if task already exist
        """
        try:
            schedule = IntervalSchedule.objects.get(every=self.interval_time, period=IntervalSchedule.SECONDS)
        except ProgrammingError:
            return
        except IntervalSchedule.DoesNotExist:
            schedule = IntervalSchedule.objects.create(every=self.interval_time, period=IntervalSchedule.SECONDS)
        try:
            PeriodicTask.objects.create(interval=schedule, name=f"period_{self.name}", task=self.name)
            info = f"Period task {self.name} created"
            self.info(info)
        except IntegrityError:
            pass

    def delete_task(self) -> None:
        """Remove task from database.

        Raises:
            ValidationError: if task doesn't exist
        """
        try:
            PeriodicTask.objects.get(task=self.name).delete()
            info = f"Period task {self.name} deleted"
            self.info(info)
        except PeriodicTask.DoesNotExist:
            pass
        except ProgrammingError:
            pass

    def register_task(self) -> None:
        """Register task to database."""
        celery_app.register_task(self)
        self.create_task()
        info = f"Period task {self.name} registered"
        self.info(info)

    def active_tasks(self) -> dict:
        """Return active tasks."""
        return celery_app.control.inspect().active()

    def num_active_tasks(self) -> int:
        """Return number of active tasks."""
        active_tasks = self.active_tasks()
        count = 0
        for worker in active_tasks:
            for task in active_tasks[worker]:
                if task["name"] == self.name:
                    count += 1
        return count

    def avoid_overlap(self, *, verbose: Optional[bool] = False) -> bool:
        """Avoid task overlap.

        Args:
            verbose (bool, optional): Whether to print logs. Defaults to False.

        Returns:
            bool: True if task is not running, False otherwise
        """
        if self.num_active_tasks() > 1:
            if verbose:
                info = f"Period task {self.name} already running"
                self.info(info)
            return False
        return True



from django.db import models

from django_napse.simulations.models.simulations.managers import SimulationDataPointManager, SimulationManager
from django_napse.utils.usefull_functions import process_value_from_type


class Simulation(models.Model):
    simulation_reference = models.UUIDField(unique=True, editable=False, null=True)
    space = models.ForeignKey("django_napse_core.Space", on_delete=models.CASCADE, null=True)
    bot = models.OneToOneField("django_napse_core.Bot", on_delete=models.CASCADE, null=True, related_name="simulation")

    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    created_at = models.DateTimeField(auto_now_add=True)

    objects = SimulationManager()

    def __str__(self) -> str:
        return f"SIMULATION {self.pk}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Simulation {self.pk}:\n"
        string += f"{beacon}\t{self.bot=}\n"
        string += f"{beacon}\t{self.space=}\n"

        string += f"{beacon}\t{self.start_date=}\n"
        string += f"{beacon}\t{self.end_date=}\n"
        string += f"{beacon}\t{self.created_at=}\n"

        string += f"{beacon}Data points:\n"
        data_points = self.data_points.all().order_by("date")

        if data_points.count() > 10:
            for data_point in data_points[:5]:
                string += f"{beacon}\t{data_point.to_str()}\n"
            string += f"{beacon}\t...\n"
            for data_point in data_points[data_points.count() - 5 :]:
                string += f"{beacon}\t{data_point.to_str()}\n"
            string += f"{beacon}\t({data_points.count()} data points)\n"

        elif data_points.count() > 0:
            for data_point in data_points:
                string += f"{beacon}\t{data_point.to_str()}\n"

        else:
            string += f"{beacon}\tNo data points\n"
        if verbose:
            print(string)
        return string


class SimulationDataPoint(models.Model):
    simulation = models.ForeignKey("Simulation", on_delete=models.CASCADE, related_name="data_points")
    date = models.DateTimeField()
    value = models.FloatField()
    action = models.CharField(max_length=10)
    amount = models.FloatField()
    ticker = models.CharField(max_length=10)

    objects = SimulationDataPointManager()

    def __str__(self):
        return f"SIMULATION DATA POINT {self.pk}"

    def to_dict(self):
        extra_info = self.extra_info.all()
        return {
            "date": self.date,
            "value": self.value,
            "action": self.action,
            "amount": self.amount,
            "ticker": self.ticker,
            **{info.key: info.get_value() for info in extra_info},
        }

    def to_str(self):
        string = ""
        for key, value in self.to_dict().items():
            string += f"{key}: {value}\t"
        return string


class SimulationDataPointExtraInfo(models.Model):
    data_point = models.ForeignKey("SimulationDataPoint", on_delete=models.CASCADE, related_name="extra_info")
    key = models.CharField(max_length=64)
    value = models.CharField(max_length=64)
    target_type = models.CharField(max_length=64)

    def __str__(self):
        return f"SIMULATION DATA POINT EXTRA INFO {self.pk}"

    def get_value(self):
        return process_value_from_type(value=self.value, target_type=self.target_type)



from django_napse.api.keys import KeyView
from django_napse.utils.api_test_case import APITestCase
from django_napse.utils.dict_comparison import compare_responses

"""
python tests/test_app/manage.py test tests.django_tests.api.keys.test_create -v2 --keepdb --parallel
"""


class KeyCreateAPITestCase:
    def test_create(self):
        self.run_tests("create")

    def setup(self):
        def create_all_good():
            def setup__create():
                return self.client.post(path=self.url, data=self.data, headers=self.headers)

            def check__create(response):
                return compare_responses(response.data, {"key": ""})

            return {
                "name": "create_all_good",
                "setup": setup__create,
                "status_code": 201,
                "checks": [check__create],
            }

        def create_missing_username():
            def setup__create():
                return self.client.post(path=self.url, data={}, headers=self.headers)

            def check__create(response):
                return compare_responses(response.data, {"error": ""})

            return {
                "name": "create_missing_username",
                "setup": setup__create,
                "status_code": 400,
                "data": {},
                "checks": [check__create],
            }

        return {
            "create": {
                "url_name": "key-list",
                "view": KeyView,
                "method": "POST",
                "data": {"username": "test_username"},
                "tests": [
                    # create_all_good(),
                    # create_missing_username(),
                ],
            },
        }


class KeyCreateAPIBinanceTestCase(KeyCreateAPITestCase, APITestCase):
    exchange = "BINANCE"



class OrderError:
    """Base class for order errors."""

    class InvalidOrder(Exception):
        """Raised when a order is invalid."""

    class StatusError(Exception):
        """Raised when the status of an order is changed at the wrong time."""

    class ProcessError(Exception):
        """Raised when an error occurs during the processing of an order."""



from django_napse.core.tasks import BaseTask
from django_napse.simulations.models import SimulationQueue
from django_napse.utils.errors import SimulationError


class SimulationQueueTask(BaseTask):
    """Task to process SimulationQueues."""

    name = "simulation_queue"
    interval_time = 5
    time_limit = 60 * 60
    soft_time_limit = 60 * 60

    def run(self) -> None:
        """Run a task to process all SimulationQueues.

        Raises:
            e: If the SimulationQueue gebe
            SimulationError.BotSimQueueError: _description_
        """
        if not self.avoid_overlap(verbose=False):
            return
        self.info("Running SimQueueTask")
        queue = SimulationQueue.objects.filter(error=False).order_by("created_at").first()

        if queue is None:
            return
        if queue.status == "RUNNING":
            return
        try:
            queue.run_quicksim()
            queue = SimulationQueue.objects.get(simulation_reference=queue.simulation_reference)
        except SimulationError.CancelledSimulationError:
            self.info("BotSimQueueTask: cancelled")
            queue.delete()
            return
        except Exception as e:
            error = f"BotSimQueueTask: error {e}"
            self.error(error)
            queue.error = True
            queue.save()
            raise e from None

        if queue.is_finished():
            print("BotSimQueueTask: finished")
            queue.delete()
        else:
            error_msg = f"SimQueueTask: {queue} status {queue.status} completion {queue.completion}"
            raise SimulationError.BotSimQueueError(error_msg)


SimulationQueueTask().delete_task()
SimulationQueueTask().register_task()



from typing import ClassVar

from rest_framework import serializers

from django_napse.api.permissions.serializers import PermissionSerializer
from django_napse.auth.models import NapseAPIKey


class NapseAPIKeySerializer(serializers.ModelSerializer):
    """Serialize a NapseAPIKey instance."""

    permissions = PermissionSerializer(many=True, read_only=True)

    class Meta:  # noqa: D106
        model = NapseAPIKey
        fields: ClassVar[list[str]] = [
            "name",
            "prefix",
            "permissions",
            "is_master_key",
            "revoked",
            "description",
        ]


class NapseAPIKeySpaceSerializer(serializers.ModelSerializer):
    """Serialize a NapseAPIKey instance with space permissions."""

    class Meta:  # noqa: D106
        model = NapseAPIKey
        fields: ClassVar[list[str]] = [
            "name",
            "prefix",
            "is_master_key",
            "revoked",
            "description",
        ]

    def to_representation(self, instance: NapseAPIKey | None = None) -> dict[str, any]:
        """Convert the instance into a dictionary."""
        representation = super().to_representation(instance)
        instance = instance or self._instance
        representation["permissions"] = []
        space = self._kwargs.get("space", None)
        if space is None:
            return representation

        for permission in instance.permissions.filter(space__uuid=space):
            representation["permissions"].append(permission.permission_type)
        return representation



from django_napse.core.models import MBPPlugin
from django_napse.utils.constants import PLUGIN_CATEGORIES
from django_napse.utils.model_test_case import ModelTestCase
from tests.django_tests.db.bots.test_plugin import PluginDefaultTestCase

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_plugin_mbp -v2 --keepdb --parallel
"""


class MBPPluginTestCase(PluginDefaultTestCase):
    model = MBPPlugin

    def test_plugin_category(self):
        plugin = self.simple_create()
        self.assertEqual(plugin.plugin_category(), PLUGIN_CATEGORIES.POST_ORDER)


class MBPPluginBINANCETestCase(MBPPluginTestCase, ModelTestCase):
    exchange = "BINANCE"



from typing import TYPE_CHECKING

from django.db import models
from django.db.transaction import atomic

from django_napse.core.models.histories.wallet import WalletHistory

if TYPE_CHECKING:
    from django_napse.core.models.transactions.debit import Debit
    from django_napse.core.models.wallets.wallet import Wallet


class DebitManager(models.Manager):
    """Manager for the Debit model."""

    @atomic()
    def create(self, wallet: "Wallet", amount: float, ticker: str) -> "Debit":
        """Create a Debit object.

        Args:
            wallet (Wallet): The wallet to debit.
            amount (float): The amount to debit.
            ticker (str): The ticker of the currency to debit.

        Returns:
            Debit: The created debit.
        """
        if amount == 0:
            return None
        debit = self.model(
            wallet=wallet,
            amount=amount,
            ticker=ticker,
        )

        debit.save()
        wallet.spend(amount, ticker, force=True)
        WalletHistory.get_or_create(wallet).generate_data_point()
        return debit



# Generated by Django 4.2.5 on 2023-10-09 15:35

from django.db import migrations, models
import django.db.models.deletion
import django_napse.utils.constants
import uuid


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("django_napse_core", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Simulation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "simulation_reference",
                    models.UUIDField(editable=False, null=True, unique=True),
                ),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "bot",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="simulation",
                        to="django_napse_core.bot",
                    ),
                ),
                (
                    "space",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_napse_core.napsespace",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SimulationDataPoint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField()),
                ("value", models.FloatField()),
                ("action", models.CharField(max_length=10)),
                ("amount", models.FloatField()),
                ("ticker", models.CharField(max_length=10)),
                (
                    "simulation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data_points",
                        to="django_napse_simulations.simulation",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SimulationQueue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "simulation_reference",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField()),
                ("canceled", models.BooleanField(default=False)),
                (
                    "status",
                    models.CharField(
                        default=django_napse.utils.constants.SIMULATION_STATUS["IDLE"],
                        max_length=12,
                    ),
                ),
                ("completion", models.FloatField(default=0.0)),
                ("eta", models.DurationField(blank=True, null=True)),
                ("error", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "bot",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_napse_core.bot",
                    ),
                ),
                (
                    "space",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_napse_core.napsespace",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SimulationDataPointExtraInfo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=64)),
                ("value", models.CharField(max_length=64)),
                ("target_type", models.CharField(max_length=64)),
                (
                    "data_point",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="extra_info",
                        to="django_napse_simulations.simulationdatapoint",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DataSetQueue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "controller",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dataset_queues",
                        to="django_napse_core.controller",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DataSet",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("start_date", models.DateTimeField(blank=True, null=True)),
                ("end_date", models.DateTimeField(blank=True, null=True)),
                ("last_update", models.DateTimeField(auto_now_add=True)),
                (
                    "status",
                    models.CharField(
                        default=django_napse.utils.constants.DOWNLOAD_STATUS["IDLE"],
                        max_length=12,
                    ),
                ),
                ("completion", models.FloatField(default=0.0)),
                ("eta", models.DurationField(blank=True, null=True)),
                (
                    "controller",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dataset",
                        to="django_napse_core.controller",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SimulationQueueInvestedCurrency",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ticker", models.CharField(max_length=10)),
                ("amount", models.FloatField(default=0)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="investments",
                        to="django_napse_simulations.simulationqueue",
                    ),
                ),
            ],
            options={
                "unique_together": {("owner", "ticker")},
            },
        ),
        migrations.CreateModel(
            name="SimulationInvestedCurrency",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ticker", models.CharField(max_length=10)),
                ("amount", models.FloatField(default=0)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="investments",
                        to="django_napse_simulations.simulation",
                    ),
                ),
            ],
            options={
                "unique_together": {("owner", "ticker")},
            },
        ),
        migrations.CreateModel(
            name="Candle",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("open_time", models.DateTimeField()),
                ("open", models.FloatField()),
                ("high", models.FloatField()),
                ("low", models.FloatField()),
                ("close", models.FloatField()),
                ("volume", models.FloatField()),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="candles",
                        to="django_napse_simulations.dataset",
                    ),
                ),
            ],
            options={
                "unique_together": {("dataset", "open_time")},
            },
        ),
    ]



from django_napse.core.tasks import BaseTask
from django_napse.simulations.models import DataSetQueue
from django_napse.utils.errors import SimulationError


class DataSetQueueTask(BaseTask):
    """Task to process DataSetQueues."""

    name = "dataset_queue"
    interval_time = 5
    time_limit = 60 * 60
    soft_time_limit = 60 * 60

    def run(self) -> None:
        """Run a task to process all DataSetQueues.

        Raises:
            SimulationError.DataSetQueueError: If the DataSetQueue is not finished.
        """
        if not self.avoid_overlap(verbose=False):
            return
        self.info("Running DataSetQueueTask")
        queue = DataSetQueue.objects.all().order_by("created_at").first()
        if queue is None:
            return
        try:
            queue.download()
        except TimeoutError:
            return
        if queue.is_finished():
            queue.delete()
        else:
            dataset = queue.get_dataset()
            error_msg = f"DataSetQueueTask: {queue} status {dataset.status} completion {dataset.completion}"
            raise SimulationError.DataSetQueueError(error_msg)


DataSetQueueTask().delete_task()
DataSetQueueTask().register_task()



from django.apps import apps
from django.db import models


class ClusterManager(models.Manager):
    def create(self, fleet, share, breakpoint, autoscale):  # noqa A002
        SpecificShare = apps.get_model("django_napse_core", "SpecificShare")
        SpecificBreakPoint = apps.get_model("django_napse_core", "SpecificBreakPoint")
        SpecificAutoscale = apps.get_model("django_napse_core", "SpecificAutoscale")

        cluster = self.model(fleet=fleet)
        cluster.save()
        SpecificShare.objects.create(cluster=cluster, share=share)
        SpecificBreakPoint.objects.create(cluster=cluster, breakpoint=breakpoint)
        SpecificAutoscale.objects.create(cluster=cluster, autoscale=autoscale)



from contextlib import suppress

from django.test import TestCase

from django_napse.core.settings import napse_settings
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.test_validation -v2 --keepdb --parallel
"""


class ValidateModelTestCaseSubclassesTestCase(TestCase):
    def test_naming_convention(self):
        for subclass in ModelTestCase.__subclasses__():
            if not subclass.__name__.endswith("TestCase"):
                error_msg = f"ModelTestCase subclass {subclass.__name__} name must end with TestCase."
                error_msg += " Make sure you are writing tests as specified in the docs."
                raise ValueError(error_msg)

    def test_all_exchanges_are_tested(self):
        all_tests = {}
        for subclass in ModelTestCase.__subclasses__():
            with suppress(AttributeError):
                if subclass.skip_exchange_validation:
                    continue
            if subclass.exchange is None:
                error_msg = f"ModelTestCase subclass {subclass.__name__} must have an exchange attribute."
                error_msg += " Make sure you are writing tests as specified in the docs."
                raise ValueError(error_msg)

            if subclass.__name__.endswith(f"{subclass.exchange}TestCase"):
                subclass_name = subclass.__name__.replace(f"{subclass.exchange}TestCase", "")
                all_tests[subclass_name] = [*all_tests.get(subclass_name, []), subclass.exchange]

            else:
                error_msg = f"ModelTestCase subclass {subclass.__name__} must end with <Exchange>TestCase."
                error_msg += " Make sure you are writing tests as specified in the docs."
                raise ValueError(error_msg)

        for subclass_name, exchanges in all_tests.items():
            if exchanges != napse_settings.NAPSE_EXCHANGES_TO_TEST:
                print(exchanges, napse_settings.NAPSE_EXCHANGES_TO_TEST)
                diff = [exchange for exchange in napse_settings.NAPSE_EXCHANGES_TO_TEST if exchange not in exchanges]
                error_msg = f"ModelTestCase subclass {subclass_name} is missing tests for exchanges {diff}"
                raise ValueError(error_msg)



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models.bots.architecture import Architecture


class ArchitectureSerializer(serializers.ModelSerializer):
    """Serialize an Architecture instance."""

    class Meta:  # noqa: D106
        model = Architecture
        fields = "__all__"
        read_only_fields: ClassVar[list[str]] = [
            "id",
        ]



# Generated by Django 4.2.5 on 2023-10-05 15:54

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="History",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="HistoryDataPoint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data_points",
                        to="django_napse_core.history",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="HistoryDataPointField",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=255)),
                ("value", models.CharField(max_length=255)),
                ("target_type", models.CharField(max_length=255)),
                (
                    "history_data_point",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fields",
                        to="django_napse_core.historydatapoint",
                    ),
                ),
            ],
        ),
    ]



from typing import TYPE_CHECKING

from django.db import models
from django.db.transaction import atomic

from django_napse.core.models.histories.wallet import WalletHistory
from django_napse.utils.constants import TRANSACTION_TYPES
from django_napse.utils.errors import TransactionError

if TYPE_CHECKING:
    from django_napse.core.models.transactions.transaction import Transaction
    from django_napse.core.models.wallets.wallet import Wallet


class TransactionManager(models.Manager):
    """The manager for the Transaction model."""

    @atomic()
    def create(self, from_wallet: "Wallet", to_wallet: "Wallet", amount: float, ticker: str, transaction_type: str) -> "Transaction":
        """Create a Transaction object and update the wallets accordingly.

        Args:
            from_wallet (Wallet): The wallet to take the money from.
            to_wallet (Wallet): The wallet to send the money to.
            amount (float): The amount of money to transfer.
            ticker (str): The ticker of the currency to transfer.
            transaction_type (str): The type of transaction.

        Raises:
            TransactionError.DifferentAccountError: If the wallets are on different exchange_accounts.
            TransactionError.SameWalletError: If the wallets are the same.
            TransactionError.TestingError: If the wallets are not both testing or both not testing.
            TransactionError.InvalidTransactionError: If the transaction type is not in TRANSACTION_TYPES.

        Returns:
            Transaction: The created transaction.
        """
        if amount == 0:
            return None
        transaction = self.model(
            from_wallet=from_wallet,
            to_wallet=to_wallet,
            amount=amount,
            ticker=ticker,
            transaction_type=transaction_type,
        )
        if from_wallet.exchange_account != to_wallet.exchange_account:
            error_msg = "Wallets must be on the same exchange_account."
            raise TransactionError.DifferentAccountError(error_msg)

        if from_wallet == to_wallet:
            error_msg = "Wallets must be different."
            raise TransactionError.SameWalletError(error_msg)

        if from_wallet.testing != to_wallet.testing:
            error_msg = f"Wallets must be both testing or both not testing. Here: {from_wallet.testing} -> {to_wallet.testing}."
            raise TransactionError.TestingError(error_msg)

        if transaction_type not in TRANSACTION_TYPES:
            error_msg = f"Transaction type {transaction_type} not in {TRANSACTION_TYPES}."
            raise TransactionError.InvalidTransactionError(error_msg)

        from_wallet.spend(amount, ticker, force=True)
        to_wallet.top_up(amount, ticker, mbp=from_wallet.currencies.get(ticker=ticker).mbp, force=True)
        transaction.save()
        WalletHistory.get_or_create(from_wallet).generate_data_point()
        WalletHistory.get_or_create(to_wallet).generate_data_point()
        return transaction



from django.db.utils import IntegrityError

from django_napse.core.models import Currency
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.wallets.test_currency -v2 --keepdb --parallel
"""


class CurrencyTestCase:
    model = Currency

    def setUp(self) -> None:
        self.wallet = self.space.wallet

    def simple_create(self):
        return Currency.objects.create(wallet=self.wallet, ticker="BTC", amount=1, mbp=20000)

    def test_create_currency(self):
        Currency.objects.create(wallet=self.wallet, ticker="BTC", amount=1, mbp=20000)

    def test_duplicate_currency_for_a_wallet(self):
        Currency.objects.create(wallet=self.wallet, ticker="BTC", amount=1, mbp=20000)
        with self.assertRaises(IntegrityError):
            Currency.objects.create(wallet=self.wallet, ticker="BTC", amount=1, mbp=20000)

    def test_property_testing(self):
        currency = Currency.objects.create(wallet=self.wallet, ticker="BTC", amount=1, mbp=20000)
        self.assertTrue(currency.testing)


class CurrencyBINANCETestCase(CurrencyTestCase, ModelTestCase):
    exchange = "BINANCE"



from django.apps import apps
from django.db import models

from django_napse.utils.constants import EXCHANGE_PAIRS, ORDER_LEEWAY_PERCENTAGE, SIDES, TRANSACTION_TYPES
from django_napse.utils.errors import OrderError


class OrderManager(models.Manager):
    def create(
        self,
        batch,
        connection,
        asked_for_amount: float,
        asked_for_ticker: str,
        pair: str,
        price: float,
        side: str,
    ):
        if side not in [SIDES.BUY, SIDES.SELL, SIDES.KEEP]:
            error_msg = f"Side {side} is not valid."
            raise OrderError.InvalidOrder(error_msg)

        if side == SIDES.BUY and asked_for_ticker != batch.controller.quote:
            error_msg = f"Ticker {asked_for_ticker} is not valid for a buy order. Should be {batch.controller.quote}."
            raise OrderError.InvalidOrder(error_msg)

        if side == SIDES.SELL and asked_for_ticker != batch.controller.base:
            error_msg = f"Ticker {asked_for_ticker} is not valid for a sell order. Should be {batch.controller.base}."
            raise OrderError.InvalidOrder(error_msg)

        # if side == SIDES.KEEP:
        #     return None

        order = self.model(
            batch=batch,
            connection=connection,
            asked_for_amount=asked_for_amount,
            pair=pair,
            price=price,
            side=side,
            asked_for_ticker=asked_for_ticker,
        )
        order.save()

        OrderWallet = apps.get_model("django_napse_core", "OrderWallet")
        wallet = OrderWallet.objects.create(title=f"Wallet for order {order.pk}", owner=order)

        Transaction = apps.get_model("django_napse_core", "Transaction")
        Transaction.objects.create(
            from_wallet=connection.wallet,
            to_wallet=wallet,
            amount=asked_for_amount * (1 + ORDER_LEEWAY_PERCENTAGE / 100),
            ticker=EXCHANGE_PAIRS[connection.space.exchange_account.exchange.name][pair]["base" if side == SIDES.SELL else "quote"],
            transaction_type=TRANSACTION_TYPES.ORDER_DEPOSIT,
        )
        order.debited_amount = asked_for_amount * (1 + ORDER_LEEWAY_PERCENTAGE / 100)
        order.save()
        return order



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models import Order
from django_napse.utils.constants import SIDES


class OrderSerializer(serializers.ModelSerializer):
    """.

    {
        "side": "BUY",
        "completed": true,
        "spent": {
            "ticker": "BTC",
            "amount": 0.1,
            "price": 1,
            "value": 0.1,
        },
        "received": {
            "ticker": "MATIC",
            "amount": 995,
            "price": 0.0001,
            "value": 0.095,
        },
        "fees": {
            "ticker": "BTC",
            "amount": 5,
            "price": 0.0001,
            "value": 0.05,
        },
        "created_at": "2021-09-01T12:00:00Z",
    }
    """

    spent = serializers.SerializerMethodField()
    received = serializers.SerializerMethodField()
    fees = serializers.SerializerMethodField()

    class Meta:  # noqa: D106
        model = Order
        fields: ClassVar[list[str]] = [
            "side",
            "completed",
            "spent",
            "received",
            "fees",
            "created_at",
        ]
        read_only_fields = fields

    def get_spent(self, instance: Order) -> dict[str, str | float]:
        """Return spend informations."""
        exit_amount = instance.exit_amount_quote if instance.side == SIDES.BUY else instance.exit_amount_base
        amount = instance.debited_amount - exit_amount
        ticker = instance.tickers_info().get("spent_ticker")
        return {
            "ticker": ticker,
            "amount": amount,
            "price": 1,
            "value": amount,
        }

    def get_received(self, instance: Order) -> dict[str, str | float]:
        """Return receive informations."""
        amount = instance.exit_amount_base if instance.side == SIDES.BUY else instance.exit_amount_quote
        ticker = instance.tickers_info().get("received_ticker")
        price = instance.price
        return {
            "ticker": ticker,
            "amount": amount,
            "price": price,
            "value": amount * price,
        }

    def get_fees(self, instance: Order) -> dict[str, str | float]:
        """Return fees info about the market order."""
        return {
            "ticker": instance.fee_ticker,
            "amount": instance.fees,
            "value": instance.fees * instance.price,
        }



from django.db import models


class StrategyManager(models.Manager):
    def create(self, **kwargs):
        architecture = kwargs["architecture"]
        config = kwargs["config"]
        if architecture.__class__ != self.model.architecture_class():
            error_msg = f"architecture must be of type {self.model.architecture_class().__name__}"
            raise TypeError(error_msg)
        if config.__class__ != self.model.config_class():
            error_msg = f"config must be of type {self.model.config_class().__name__}"
            raise TypeError(error_msg)
        strategy = self.model(**kwargs)
        strategy.save()
        for plugin in self.model.plugin_classes():
            plugin.objects.create(strategy=strategy)
        return strategy



class BotError:
    """Base class for bot errors."""

    class InvalidSetting(Exception):
        """Raised when a bot is invalid."""

    class NoSpace(Exception):
        """Raised when a bot is in a fleet and therefore doesn't have a space."""

    class BuildNotPossible(Exception):
        """Raised when a bot cannot be built due to bad parameters."""

    class InvalidTicker(Exception):
        """Raised when a ticker is invalid."""


class BotConfigError:
    """Base class for bot config errors."""

    class MissingSettingError(Exception):
        """Raised when a setting is missing."""

    class DuplicateBotConfig(Exception):
        """Raised if an identical BotConfig already exists in this space."""


class ControllerError:
    """Base class for controller errors."""

    class BypassError(Exception):
        """Raised when a bypass isn't used."""

    class InvalidSetting(Exception):
        """Raised when a controller is invalid."""

    class InvalidTicker(Exception):
        """Raised when a ticker is invalid."""

    class InvalidPair(Exception):
        """Raised when a pair is invalid."""


class ClusterError:
    """Base class for cluster errors."""

    class MutableBotConfig(Exception):
        """Raised when a BotConfig in a cluster is mutable."""

    class MultipleConfigs(Exception):
        """Raised when a cluster has multiple configs."""


class PluginError:
    """Base class for plugin errors."""

    class InvalidPluginCategory(Exception):
        """Raised when a plugin category is invalid."""



from __future__ import annotations

from typing import TYPE_CHECKING

from rest_framework.viewsets import GenericViewSet

from django_napse.auth.models import NapseAPIKey
from django_napse.core.models import Space
from django_napse.utils.errors import APIError

if TYPE_CHECKING:
    from rest_framework.request import Request


class CustomViewSet(GenericViewSet):
    """Base of all ViewSets."""

    def get_api_key(self, request: Request) -> NapseAPIKey:
        """Return the api key from the request."""
        try:
            return NapseAPIKey.objects.get_from_key(request.META["HTTP_AUTHORIZATION"].split()[1])
        except NapseAPIKey.DoesNotExist as e:
            raise APIError.InvalidAPIKey from e
        except KeyError as e:
            raise APIError.NoAPIKey from e

    def get_space(self, request: Request) -> Space | None:
        """Return the space from the request."""
        try:
            return Space.objects.get(uuid=request.query_params["space"])
        except Space.DoesNotExist:
            return None



import uuid
from datetime import datetime, timedelta

import pandas as pd
from django.db import models
from django.utils.timezone import get_default_timezone

from django_napse.core.models.histories.managers.history_data_point import HistoryDataPointManager
from django_napse.core.models.histories.managers.history_data_point_field import HistoryDataPointFieldManager
from django_napse.utils.constants import HISTORY_DATAPOINT_FIELDS, HISTORY_DATAPOINT_FIELDS_WILDCARDS
from django_napse.utils.errors import HistoryError
from django_napse.utils.findable_class import FindableClass
from django_napse.utils.usefull_functions import process_value_from_type


class History(FindableClass, models.Model):
    """A History is a collection of data points.

    Use it to track the evolution of a value over time.

    Create a child class to get started.

    The child class should have a ForeignKey to the model you want to track (called `owner`).
    """

    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)

    def __str__(self) -> str:
        return f"HISTORY {self.uuid}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}History {self.pk}:\n"
        string += f"{beacon}\t{self.uuid=}\n"
        string += f"{beacon}\tDataPoints:\n"
        df_string = str(self.to_dataframe()).replace("\n", f"\n{beacon}\t\t")
        string += f"{beacon}\t\t{df_string}\n"
        if verbose:
            print(string)
        return string

    def to_dataframe(self) -> pd.DataFrame:
        """Return a DataFrame containing the data points."""
        all_data_points = self.data_points.all()
        return pd.DataFrame([data_point.to_dict() for data_point in all_data_points])

    @property
    def owner(self) -> models.Model:
        """Return the owner of the history."""
        return self.find().owner

    @classmethod
    def get_or_create(cls, owner: models.Model) -> "History":
        """Return the history of the owner if it exists, else create it."""
        if hasattr(owner, "history"):
            return owner.history
        return cls.objects.create(owner=owner)

    def delta(self, days: int = 30) -> float:
        """Return the value delta between today and {days} days ago."""
        date = datetime.now(tz=get_default_timezone()) - timedelta(days=days)
        data_points = self.data_points.filter(created_at__date=date.date())

        while not data_points.exists():
            days -= 1
            date = date + timedelta(days=days)
            data_points = self.data_points.filter(created_at__date=date.date())
            if days == 0 and not data_points.exists():
                return 0
            if data_points.exists():
                break

        return data_points

    def generate_data_point(self) -> "HistoryDataPoint":
        """Create a new data point.

        This method should be implemented in the child class.
        """
        error_msg = "You must implement the generate_data_point method in your child class."
        raise NotImplementedError(error_msg)


class HistoryDataPoint(models.Model):
    """A HistoryDataPoint is a collection of fields."""

    history: "History" = models.ForeignKey(History, on_delete=models.CASCADE, related_name="data_points")
    created_at = models.DateTimeField(auto_now_add=True)

    objects = HistoryDataPointManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"HISTORY DATA POINT {self.pk} {self.history.uuid}"

    def to_dict(self) -> dict:
        """Return a dictionary containing the fields."""
        return {field.key: field.get_value() for field in self.fields.all()}

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}HistoryDataPoint {self.pk}:\n"
        string += f"{beacon}Fields:\n"
        for field in self.fields.all():
            string += field.info(beacon=beacon + "\t", verbose=False)
        if verbose:
            print(string)
        return string


class HistoryDataPointField(models.Model):
    """A HistoryDataPointField is a key-value pair with a target type."""

    history_data_point: "HistoryDataPoint" = models.ForeignKey(HistoryDataPoint, on_delete=models.CASCADE, related_name="fields")
    key = models.CharField(max_length=255)
    value = models.CharField(max_length=255)
    target_type = models.CharField(max_length=255)

    objects = HistoryDataPointFieldManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"HISTORY DATA POINT FIELD {self.pk} {self.history_data_point.pk}"

    def save(self, *args, **kwargs):  # noqa
        is_wildcard = False
        for wildcard in HISTORY_DATAPOINT_FIELDS_WILDCARDS:
            if self.key.startswith(wildcard):
                is_wildcard = True
                break
        if self.key not in HISTORY_DATAPOINT_FIELDS and not is_wildcard:
            error_msg = f"Invalid key {self.key} for HistoryDataPointField"
            raise HistoryError.InvalidDataPointFieldKey(error_msg)
        return super().save(*args, **kwargs)

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}HistoryDataPointField {self.pk}:\n"
        string += f"{beacon}\t{self.key=}\n"
        string += f"{beacon}\t{self.value=}\n"
        string += f"{beacon}\t{self.target_type=}\n"
        if verbose:
            print(string)
        return string

    def get_value(self) -> any:
        """Return the value as the target type."""
        return process_value_from_type(value=self.value, target_type=self.target_type)



import math

from django.db import models

from django_napse.core.models.bots.architectures.single_pair import SinglePairArchitecture
from django_napse.core.models.bots.implementations.turbo_dca.config import TurboDCABotConfig
from django_napse.core.models.bots.plugins import LBOPlugin, MBPPlugin, SBVPlugin
from django_napse.core.models.bots.strategy import Strategy
from django_napse.core.models.wallets.currency import CurrencyPydantic
from django_napse.utils.constants import SIDES


class TurboDCAStrategy(Strategy):
    variable_last_buy_date = models.DateTimeField(null=True, blank=True)

    def __str__(self) -> str:
        return f"TURBO DCA BOT STRATEGY: {self.pk=}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Strategy ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.config=}\n"
        string += f"{beacon}\t{self.architecture=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @classmethod
    def config_class(cls):
        return TurboDCABotConfig

    @classmethod
    def architecture_class(cls):
        return SinglePairArchitecture

    @classmethod
    def plugin_classes(cls):
        return [MBPPlugin, LBOPlugin, SBVPlugin]

    def give_order(self, data: dict) -> list[dict]:
        controller = data["controllers"]["main"]
        if (
            self.variable_last_buy_date is None
            or data["candles"][controller]["current"]["open_time"] - self.variable_last_buy_date >= data["config"]["timeframe"]
        ):
            mbp = data["connection_data"][data["connection"]]["connection_specific_args"]["mbp"].get_value()
            lbo = data["connection_data"][data["connection"]]["connection_specific_args"]["lbo"].get_value()
            sbv = data["connection_data"][data["connection"]]["connection_specific_args"]["sbv"].get_value()
            available_base = (
                data["connection_data"][data["connection"]]["wallet"]
                .currencies.get(controller.base, CurrencyPydantic(ticker=controller.base, amount=0, mbp=0))
                .amount
            )
            available_quote = (
                data["connection_data"][data["connection"]]["wallet"]
                .currencies.get(controller.quote, CurrencyPydantic(ticker=controller.quote, amount=0, mbp=0))
                .amount
            )
            mbp = mbp if mbp is not None else math.inf
            sbv = sbv if sbv is not None else available_quote
            current_price = data["candles"][controller]["latest"]["close"]
            amount = data["config"]["percentage"] * sbv / 100
            if lbo == 0 or current_price < mbp:
                return [
                    {
                        "controller": controller,
                        "ArchitectureModifications": [],
                        "StrategyModifications": [
                            {
                                "key": "last_buy_date",
                                "value": str(data["candles"][controller]["current"]["open_time"]),
                                "target_type": "datetime",
                                "ignore_failed_order": False,
                            },
                        ],
                        "ConnectionModifications": [],
                        "connection": data["connection"],
                        "asked_for_amount": amount,
                        "asked_for_ticker": controller.quote,
                        "pair": controller.pair,
                        "price": data["candles"][controller]["latest"]["close"],
                        "side": SIDES.BUY,
                    },
                ]
            if current_price > mbp:
                return [
                    {
                        "controller": controller,
                        "ArchitectureModifications": [],
                        "StrategyModifications": [
                            {
                                "key": "last_buy_date",
                                "value": str(data["candles"][controller]["current"]["open_time"]),
                                "target_type": "datetime",
                                "ignore_failed_order": False,
                            },
                        ],
                        "ConnectionModifications": [],
                        "connection": data["connection"],
                        "asked_for_amount": available_base,
                        "asked_for_ticker": controller.base,
                        "pair": controller.pair,
                        "price": data["candles"][controller]["latest"]["close"],
                        "side": SIDES.SELL,
                    },
                ]
            return [
                {
                    "controller": controller,
                    "ArchitectureModifications": [],
                    "StrategyModifications": [
                        {
                            "key": "last_buy_date",
                            "value": str(data["candles"][controller]["current"]["open_time"]),
                            "target_type": "datetime",
                            "ignore_failed_order": False,
                        },
                    ],
                    "ConnectionModifications": [],
                    "connection": data["connection"],
                    "asked_for_amount": 0,
                    "asked_for_ticker": controller.quote,
                    "pair": controller.pair,
                    "price": data["candles"][controller]["latest"]["close"],
                    "side": SIDES.KEEP,
                },
            ]
        return [
            {
                "controller": controller,
                "ArchitectureModifications": [],
                "StrategyModifications": [],
                "ConnectionModifications": [],
                "connection": data["connection"],
                "asked_for_amount": 0,
                "asked_for_ticker": controller.quote,
                "pair": controller.pair,
                "price": data["candles"][controller]["latest"]["close"],
                "side": SIDES.KEEP,
            },
        ]



# Generated by Django 4.2.6 on 2023-10-23 19:44

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0006_alter_exchangeaccount_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="historydatapoint",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
    ]



from django.forms import ValidationError
from rest_framework.permissions import BasePermission
from rest_framework_api_key.permissions import HasAPIKey  # noqa

from django_napse.core.models import Space
from django_napse.utils.constants import PERMISSION_TYPES
from django_napse.utils.errors import APIError


def check_for_space(request):
    if "space" not in request.query_params:
        raise APIError.MissingSpace()
    try:
        return Space.objects.get(uuid=request.query_params["space"])
    except Space.DoesNotExist as e:
        raise APIError.InvalidSpace() from e
    except ValidationError as e:
        raise APIError.InvalidSpace() from e


class HasAdminPermission(BasePermission):
    def has_permission(self, request, view):
        space = check_for_space(request)

        api_key = view.get_api_key(request)
        if api_key.is_master_key:
            return True
        if any(
            permission.permission_type == PERMISSION_TYPES.ADMIN
            for permission in api_key.permissions.filter(
                space=space,
                approved=True,
                revoked=False,
            )
        ):
            return True
        raise APIError.InvalidPermissions()


class HasFullAccessPermission(BasePermission):
    def has_permission(self, request, view):
        space = check_for_space(request)

        api_key = view.get_api_key(request)
        if api_key.is_master_key:
            return True
        for permission in api_key.permissions.filter(
            space=space,
            approved=True,
            revoked=False,
        ):
            if permission.permission_type in [PERMISSION_TYPES.ADMIN, PERMISSION_TYPES.FULL_ACCESS]:
                return True
        raise APIError.InvalidPermissions()


class HasReadPermission(BasePermission):
    def has_permission(self, request, view):
        space = check_for_space(request)

        api_key = view.get_api_key(request)
        if api_key.is_master_key:
            return True
        for permission in api_key.permissions.filter(
            space=space,
            approved=True,
            revoked=False,
        ):
            if permission.permission_type in [PERMISSION_TYPES.ADMIN, PERMISSION_TYPES.FULL_ACCESS, PERMISSION_TYPES.READ_ONLY]:
                return True
        raise APIError.InvalidPermissions()


class HasSpace(BasePermission):
    def has_permission(self, request, view):
        check_for_space(request)
        return True


class HasMasterKey(BasePermission):
    def has_permission(self, request, view):
        api_key = view.get_api_key(request)
        if api_key.is_master_key:
            return True
        raise APIError.InvalidPermissions()



import time
import uuid
from contextlib import suppress
from copy import deepcopy
from datetime import datetime, timedelta

from django.db import models

from django_napse.core.models.bots.bot import Bot
from django_napse.core.models.bots.controller import Controller
from django_napse.core.models.modifications import ArchitectureModification, ConnectionModification, StrategyModification
from django_napse.core.models.orders.order import Order, OrderBatch
from django_napse.core.models.transactions.credit import Credit
from django_napse.core.models.wallets.currency import CurrencyPydantic
from django_napse.simulations.models.datasets.dataset import Candle, DataSet
from django_napse.simulations.models.simulations.managers import SimulationQueueManager
from django_napse.utils.constants import EXCHANGE_INTERVALS, ORDER_LEEWAY_PERCENTAGE, ORDER_STATUS, SIDES, SIMULATION_STATUS
from django_napse.utils.errors import ControllerError

from .simulation import Simulation


class SimulationQueue(models.Model):
    """Queue wrapper for a simulation."""

    simulation_reference = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    space = models.ForeignKey("django_napse_core.Space", on_delete=models.CASCADE, null=True)
    bot = models.OneToOneField("django_napse_core.Bot", on_delete=models.CASCADE, null=True)

    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    canceled = models.BooleanField(default=False)

    status = models.CharField(max_length=12, default=SIMULATION_STATUS.IDLE)
    completion = models.FloatField(default=0.0)
    eta = models.DurationField(blank=True, null=True)

    error = models.BooleanField(default=False)

    created_at = models.DateTimeField(auto_now_add=True)

    objects = SimulationQueueManager()

    def __str__(self) -> str:
        return f"BOT SIM QUEUE {self.pk}"

    def info(self, verbose: bool = True, beacon: str = "") -> str:  # noqa: FBT002, FBT001
        """Return info on SimulationQueue."""
        string = ""
        string += f"{beacon}SimulationQueue {self.pk}:\n"
        string += f"{beacon}\t{self.bot=}\n"
        string += f"{beacon}\t{self.space=}\n"

        string += f"{beacon}\t{self.start_date=}\n"
        string += f"{beacon}\t{self.end_date=}\n"
        string += f"{beacon}\t{self.created_at=}\n"

        string += f"{beacon}Investments:\n"
        investments = self.investments.all()
        if investments.count() > 0:
            for investment in investments:
                string += f"{beacon}\t{investment}\n"
        else:
            string += f"{beacon}\tNo investments\n"
        if verbose:
            print(string)
        return string

    def setup_simulation(self) -> tuple[Bot, dict[str, any]]:
        """Setup investment and connections for the simulation."""
        self.space.simulation_wallet.find().reset()
        for investment in self.investments.all():
            Credit.objects.create(
                wallet=self.space.simulation_wallet,
                ticker=investment.ticker,
                amount=investment.amount,
            )
        new_bot = self.bot.copy()
        connection = new_bot.connect_to_wallet(self.space.simulation_wallet)
        for investment in self.investments.all():
            connection.deposit(investment.ticker, investment.amount)
        no_db_data = new_bot.architecture.prepare_db_data()
        return new_bot, no_db_data

    def cleanup_simulation(self, bot: Bot) -> None:
        """Reset the simulation: reset wallet & deactivate the bot."""
        self.space.simulation_wallet.reset()
        bot.hibernate()

    def preparation(self, bot, no_db_data):
        intervals = {controller.interval for controller in bot.controllers.values()}
        for interval in EXCHANGE_INTERVALS[next(iter(bot.controllers.values())).exchange_account.exchange.name]:
            if interval in intervals:
                min_interval = interval
                break

        currencies = next(iter(no_db_data["connection_data"].values()))["wallet"].currencies

        exchange_controllers = {controller: controller.exchange_controller for controller in bot.controllers.values()}

        data = {}
        datasets = []
        for controller in bot.controllers.values():
            datasets.append(DataSet.objects.create(controller=controller, start_date=self.start_date, end_date=self.end_date))
            with suppress(ControllerError.InvalidSetting):
                datasets.append(
                    DataSet.objects.create(
                        controller=Controller.get(
                            exchange_account=controller.exchange_account,
                            base=controller.base,
                            quote="USDT",
                            interval=min_interval,
                        ),
                        start_date=self.start_date,
                        end_date=self.end_date,
                    ),
                )
            with suppress(ControllerError.InvalidSetting):
                datasets.append(
                    DataSet.objects.create(
                        controller=Controller.get(
                            exchange_account=controller.exchange_account,
                            base=controller.quote,
                            quote="USDT",
                            interval=min_interval,
                        ),
                        start_date=self.start_date,
                        end_date=self.end_date,
                    ),
                )
        datasets = set(datasets)
        candles = Candle.objects.filter(dataset__in=datasets).order_by("open_time")
        open_times = candles.values_list("open_time", flat=True).distinct()
        for open_time in open_times:
            data[open_time] = {}
            for controller in bot.controllers.values():
                data[open_time][controller] = None
        for candle in candles:
            candle_dict = candle.to_dict()
            controller = candle_dict.pop("controller")
            data[candle.open_time][controller] = candle_dict

        for controller in bot.controllers.values():
            last_candle = None
            for open_time in open_times:
                if data[open_time][controller] is None:
                    data[open_time][controller] = last_candle
                else:
                    last_candle = data[open_time][controller]
        return data, currencies, exchange_controllers, min_interval

    def process_candle_data(self, candle_data, min_interval):
        processed_data = {"candles": {}, "extras": {}}
        current_prices = {}
        for controller, candle in candle_data.items():
            processed_data["candles"][controller] = {"current": candle, "latest": candle}
            if controller.quote == "USDT" and controller.interval == min_interval:
                price = candle["close"]
                current_prices[f"{controller.base}_price"] = price
        return processed_data, current_prices

    def append_data(
        self,
        connection_specific_args: dict,
        candle_data: dict,
        current_prices: dict,
        currencies: dict,
        currencies_before: dict,
        all_orders: list,
        date: datetime,
        dates: list,
        values: list,
        actions: list,
        prices: dict,
        total_amounts: dict,
        amounts: list,
        tickers: list,
        taxes: list,
        extras: dict,
    ):
        current_amounts = {}
        for controller in candle_data:
            current_amounts[f"{controller.base}_amount"] = currencies.get(
                controller.base,
                CurrencyPydantic(ticker=controller.base, amount=0, mbp=0),
            ).amount
            current_amounts[f"{controller.quote}_amount"] = currencies.get(
                controller.quote,
                CurrencyPydantic(ticker=controller.quote, amount=0, mbp=0),
            ).amount

        wallet_value = 0
        for ticker, currency in currencies.items():
            amount = currency.amount
            price = 1 if ticker == "USDT" else current_prices[f"{ticker}_price"]
            wallet_value += amount * price

        wallet_value_before = 0
        for ticker, currency in currencies_before.items():
            amount = currency.amount
            price = 1 if ticker == "USDT" else current_prices[f"{ticker}_price"]
            wallet_value_before += amount * price

        for index, order in enumerate(all_orders):
            dates.append(date + index * timedelta(seconds=1))
            values.append(round(wallet_value, 5))
            actions.append(order.side)
            for ticker_price in current_prices:
                prices[ticker_price] = [*prices.get(ticker_price, []), 1 if ticker_price == "USDT" else round(current_prices[ticker_price], 5)]
            for ticker_amount in current_amounts:
                total_amounts[ticker_amount] = [*total_amounts.get(ticker_amount, []), round(current_amounts[ticker_amount], 5)]
            taxes.append(round(order.fees * (current_prices[f"{order.fee_ticker}_price"] if order.fee_ticker != "USDT" else 1), 5))
            amounts.append(round(order.asked_for_amount, 5))
            tickers.append(order.asked_for_ticker)
            for plugin in extras:
                extras[plugin] = [*extras[plugin], connection_specific_args[plugin].get_value()]

    def quick_simulation(self, bot, no_db_data, verbose=True):
        data, currencies, exchange_controllers, min_interval = self.preparation(bot, no_db_data)
        _time = time.time()
        tpi = []
        dates = []
        values = []
        actions = []
        prices = {}
        total_amounts = {}
        taxes = []
        amounts = []
        tickers = []
        extras = {csa.key: [] for csa in next(iter(no_db_data["connection_data"].values()))["connection_specific_args"].values()}
        for date, candle_data in data.items():
            currencies_before = deepcopy(currencies)
            processed_data, current_prices = self.process_candle_data(
                candle_data=candle_data,
                min_interval=min_interval,
            )
            orders = bot._get_orders(data=processed_data, no_db_data=no_db_data)
            batches = {}
            for order in orders:
                debited_amount = order["asked_for_amount"] * (1 + ORDER_LEEWAY_PERCENTAGE / 100)
                if debited_amount > 0:
                    currencies[order["asked_for_ticker"]].amount -= debited_amount
                order["debited_amount"] = debited_amount

                controller = order["controller"]
                batches[controller] = OrderBatch(controller=controller)

            for batch in batches.values():
                batch.status = ORDER_STATUS.READY

            all_orders = []
            for controller, batch in batches.items():
                all_modifications = []
                controller_orders = [order for order in orders if order["controller"] == controller]
                order_objects = []
                for order in controller_orders:
                    order.pop("controller")
                    strategy_modifications = order.pop("StrategyModifications")
                    connection_modifications = order.pop("ConnectionModifications")
                    architecture_modifications = order.pop("ArchitectureModifications")
                    order = Order(batch=batch, **order)
                    order_objects.append(order)
                    for modification in strategy_modifications:
                        all_modifications.append(StrategyModification(order=order, **modification))
                    for modification in connection_modifications:
                        all_modifications.append(ConnectionModification(order=order, **modification))
                    for modification in architecture_modifications:
                        all_modifications.append(ArchitectureModification(order=order, **modification))

                orders = controller.process_orders(
                    no_db_data={
                        "buy_orders": [order for order in order_objects if order.side == SIDES.BUY],
                        "sell_orders": [order for order in order_objects if order.side == SIDES.SELL],
                        "keep_orders": [order for order in order_objects if order.side == SIDES.KEEP],
                        "batches": [batch],
                        "exchange_controller": exchange_controllers[controller],
                        "min_trade": controller.min_notional / processed_data["candles"][controller]["latest"]["close"],
                        "price": processed_data["candles"][controller]["latest"]["close"],
                    },
                    testing=True,
                )
                for order in orders:
                    order.apply_modifications__no_db(
                        batch=batch,
                        modifications=[modification for modification in all_modifications if modification.order == order],
                        strategy=no_db_data["strategy"],
                        architecture=no_db_data["architecture"],
                        currencies=currencies,
                    )

                    currencies[controller.base] = currencies.get(controller.base, CurrencyPydantic(ticker=controller.base, amount=0, mbp=0))
                    currencies[controller.quote] = currencies.get(controller.quote, CurrencyPydantic(ticker=controller.quote, amount=0, mbp=0))
                    currencies[controller.base].amount += order.exit_amount_base
                    currencies[controller.quote].amount += order.exit_amount_quote

                all_orders += orders

            self.append_data(
                connection_specific_args=next(iter(no_db_data["connection_data"].values()))["connection_specific_args"],
                candle_data=candle_data,
                current_prices=current_prices,
                currencies=currencies,
                currencies_before=currencies_before,
                all_orders=all_orders,
                date=date,
                dates=dates,
                values=values,
                actions=actions,
                prices=prices,
                total_amounts=total_amounts,
                taxes=taxes,
                amounts=amounts,
                tickers=tickers,
                extras=extras,
            )
            tpi.append(time.time() - _time)
            _time = time.time()

        if verbose:
            print(f"Simulation ended.\nAverage TPI: {sum(tpi) / len(tpi)}")

        return Simulation.objects.create(
            space=self.space,
            bot=bot,
            start_date=self.start_date,
            end_date=self.end_date,
            simulation_reference=self.simulation_reference,
            data={
                "dates": dates,
                "values": values,
                "actions": actions,
                "taxes": taxes,
                "amounts": amounts,
                "tickers": tickers,
                **prices,
                **total_amounts,
                **extras,
            },
        )

    def irl_simulation(self, bot, no_db_data, verbose=True):
        data, _, exchange_controllers, min_interval = self.preparation(bot, no_db_data)
        _time = time.time()
        tpi = []
        dates = []
        values = []
        actions = []
        prices = {}
        total_amounts = {}
        taxes = []
        amounts = []
        tickers = []
        extras = {csa.key: [] for csa in next(iter(no_db_data["connection_data"].values()))["connection_specific_args"].values()}
        currencies = bot.connections.all()[0].wallet.to_dict().currencies
        for date, candle_data in data.items():
            currencies_before = deepcopy(currencies)
            processed_data, current_prices = self.process_candle_data(
                candle_data=candle_data,
                min_interval=min_interval,
            )

            orders, batches = bot.get_orders(data=processed_data)

            all_orders = []
            for controller, batch in batches.items():
                orders = controller.process_orders(
                    no_db_data={
                        "buy_orders": [order for order in orders if order.side == SIDES.BUY],
                        "sell_orders": [order for order in orders if order.side == SIDES.SELL],
                        "keep_orders": [order for order in orders if order.side == SIDES.KEEP],
                        "batches": [batch],
                        "exchange_controller": exchange_controllers[controller],
                        "min_trade": controller.min_notional / processed_data["candles"][controller]["latest"]["close"],
                        "price": processed_data["candles"][controller]["latest"]["close"],
                    },
                    testing=True,
                )

                controller.apply_orders(orders)
                for order in orders:
                    order.apply_modifications()
                    order.process_payout()
                all_orders += orders

            currencies = bot.connections.all()[0].wallet.to_dict().currencies
            self.append_data(
                connection_specific_args=bot.connections.all()[0].to_dict()["connection_specific_args"],
                candle_data=candle_data,
                current_prices=current_prices,
                currencies_before=currencies_before,
                currencies=currencies,
                all_orders=all_orders,
                date=date,
                dates=dates,
                values=values,
                actions=actions,
                prices=prices,
                total_amounts=total_amounts,
                taxes=taxes,
                amounts=amounts,
                tickers=tickers,
                extras=extras,
            )
            tpi.append(time.time() - _time)
            _time = time.time()

        if verbose:
            print(f"Simulation ended.\nAverage TPI: {sum(tpi) / len(tpi)}")

        return Simulation.objects.create(
            space=self.space,
            bot=bot,
            start_date=self.start_date,
            end_date=self.end_date,
            simulation_reference=self.simulation_reference,
            data={
                "dates": dates,
                "values": values,
                "actions": actions,
                "taxes": taxes,
                "amounts": amounts,
                "tickers": tickers,
                **prices,
                **total_amounts,
                **extras,
            },
        )

    def run_quick_simulation(self, verbose=True):
        self.status = SIMULATION_STATUS.RUNNING
        self.save()
        bot, no_db_data = self.setup_simulation()

        simulation = self.quick_simulation(bot=bot, no_db_data=no_db_data, verbose=verbose)

        self.cleanup_simulation(bot)
        self.status = SIMULATION_STATUS.IDLE
        self.save()
        return simulation

    def run_irl_simulation(self, verbose=True):
        self.status = SIMULATION_STATUS.RUNNING
        self.save()
        bot, no_db_data = self.setup_simulation()

        simulation = self.irl_simulation(bot=bot, no_db_data=no_db_data, verbose=verbose)

        self.cleanup_simulation(bot)
        self.status = SIMULATION_STATUS.IDLE
        self.save()
        return simulation

    def is_finished(self):
        return self.status == SIMULATION_STATUS.IDLE and self.completion == 100

    def get_status(self):
        """Return a dictionary with the status of the BotSimQueue.

        Dict shape:
        ```json
        {
            "status": "RUNNING",
            "completion": 56.9,
            "eta": 00:01:04,
            "position_in_queue": 0,
            "error": false,
        }
        ```
        """
        return {
            "status": self.status,
            "completion": self.completion,
            "eta": self.eta,
            "position_in_queue": SimulationQueue.objects.filter(created_at__lt=self.created_at, error=False).count(),
            "error": self.error,
        }

    def cancel(self):
        """Stop the BotSim."""
        self.canceled = True
        self.save()



import math
from datetime import datetime, timedelta

from pytz import UTC


def calculate_mbp(value: str, current_value: float, order, currencies: dict) -> float:
    from django_napse.core.models.wallets.currency import CurrencyPydantic

    ticker, price = value.split("|")
    price = float(price)

    current_amount = currencies.get(ticker, CurrencyPydantic(ticker=ticker, amount=0, mbp=0)).amount
    current_value = current_value if current_value is not None else 0
    received_quote = order.debited_amount - order.exit_amount_quote
    return (current_amount * current_value + received_quote) / (received_quote / price + current_amount)


def process_value_from_type(value, target_type, **kwargs) -> any:
    """Convert a value to a specific type."""
    target_type = target_type.lower()
    if value == "None":
        return None
    if target_type == "int":
        return int(value)
    if target_type == "float":
        return float(value)
    if target_type == "bool":
        match value:
            case "True" | "true":
                return True
            case "False" | "false":
                return False
            case _:
                return value
    elif target_type == "str":
        return str(value)
    elif target_type == "datetime":
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S%z").astimezone(UTC)
    elif target_type == "date":
        return datetime.strptime(value, "%Y-%m-%d").astimezone(UTC).date()
    elif target_type == "time":
        return datetime.strptime(value, "%H:%M:%S").astimezone(UTC).time()
    elif target_type == "timedelta":
        split = value.split(", ")
        days = split[0] if len(split) == 2 else "0 days"
        timestamp = split[1] if len(split) == 2 else split[0]
        days = int(days.split(" ")[0])
        t = datetime.strptime(timestamp, "%H:%M:%S").astimezone()
        return timedelta(days=days, hours=t.hour, minutes=t.minute, seconds=t.second)
    elif target_type == "None":
        return None
    elif target_type == "plugin_mbp":
        return calculate_mbp(
            value=value,
            current_value=kwargs["current_value"],
            order=kwargs["order"],
            currencies=kwargs["currencies"],
        )
    else:
        error_message = f"Unknown target_type: {target_type}"
        raise ValueError(error_message)


def round_up(number: float, decimals: int = 0) -> float:
    """Round a number up to a given number of decimals.

    Args:
    ----
    number (float): The number to round up.
    decimals (int, optional): The number of decimals to round up to. Defaults to 0.

    Returns:
    -------
    float: The rounded number.
    """
    multiplier = 10**decimals
    return math.ceil(number * multiplier) / multiplier


def round_down(number: float, decimals: int = 0) -> float:
    """Round a number down to a given number of decimals.

    Args:
    ----
    number (float): The number to round down.
    decimals (int, optional): The number of decimals to round down to. Defaults to 0.

    Returns:
    -------
    float: The rounded number.
    """
    multiplier = 10**decimals
    return math.floor(number * multiplier) / multiplier



class FleetError:
    """Base class for fleet errors."""

    class FleetNotSetupError(Exception):
        """Raised when a fleet is not setup, but you ask I to start trading or scaling."""

    class InvalidOperator(Exception):
        """Raised when the operator is invalid."""

    class InvalidShares(Exception):
        """Raised when the sum of all shares is not 1."""

    class InvalidClusters(Exception):
        """Raised when the fleet has no clusters."""



from typing import ClassVar

from rest_framework import serializers

from django_napse.api.bots.serializers import BotSerializer
from django_napse.api.fleets.serializers.cluster_serialisers import ClusterFormatterSerializer
from django_napse.core.models import ConnectionWallet, Fleet, FleetHistory, Space, SpaceWallet


class FleetSerializer(serializers.ModelSerializer):
    """Serialize fleet instance.

    Can be use with many fleet instances.
    """

    value = serializers.SerializerMethodField(read_only=True)
    bot_count = serializers.SerializerMethodField(read_only=True)
    clusters = ClusterFormatterSerializer(
        write_only=True,
        many=True,
        required=True,
    )
    space = serializers.UUIDField(write_only=True, required=True)
    delta = serializers.SerializerMethodField(read_only=True)
    exchange_account = serializers.CharField(source="exchange_account.uuid", read_only=True)

    class Meta:  # noqa: D106
        model = Fleet
        fields: ClassVar[list[str]] = [
            "name",
            # write-only
            "clusters",
            "space",
            # read-only
            "uuid",
            "value",
            "bot_count",
            "delta",
            "exchange_account",
            "testing",
        ]
        read_only_fields: ClassVar[list[str]] = [
            "uuid",
            "value",
            "bot_count",
            "delta",
            "exchange_account",
            "testing",
        ]

    def __init__(
        self,
        *args: list[any],
        space: Space | None = None,
        **kwargs: dict[str, any],
    ) -> None:
        """."""
        self.space = space
        super().__init__(*args, **kwargs)

    def get_value(self, instance: Fleet) -> float:
        """Return fleet's value, can be space contained if space is given in serializer."""
        if self.space is None:
            return instance.value
        return instance.space_frame_value(space=self.space)

    def get_bot_count(self, instance: Fleet) -> int:
        """Return fleet's bot count, can be space contained if space is given in serializer."""
        return instance.bot_count(space=self.space)

    def get_delta(self, instance: Fleet) -> float:
        """Delta on the last 30 days."""
        try:
            history = FleetHistory.objects.get(owner=instance)
        except FleetHistory.DoesNotExist:
            return 0
        return history.get_delta()

    def validate(self, attrs: dict[str, any]) -> dict[str, any]:
        """Validation process for Fleet creation."""
        data = super().validate(attrs)
        # Space is required only to make the 1st automatic empty invest when creating a fleet (logic is in the view)
        if "space" not in attrs:
            error_msg: str = "Space is required."
            raise serializers.ValidationError(error_msg)

        try:
            self.space = Space.objects.get(uuid=attrs.pop("space"))
        except Space.DoesNotExist:
            error_msg: str = "Space does not exist."
            raise serializers.ValidationError(error_msg) from None

        data["exchange_account"] = self.space.exchange_account
        return data

    def to_representation(self, instance: Fleet) -> dict[str, any]:
        """Serialize Fleet instance."""
        data = super().to_representation(instance)
        if self.space is not None:
            data["space"] = self.space.uuid
        return data

    def create(self, validated_data: dict[str, any]) -> Fleet:
        """Create a Fleet instance."""
        return Fleet.objects.create(**validated_data)


class FleetDetailSerializer(serializers.ModelSerializer):
    """Deep serialization of a fleet instance.

    Use this serializer for endpoints details only.
    """

    wallet = serializers.SerializerMethodField(read_only=True)
    statistics = serializers.SerializerMethodField(read_only=True)
    bots = BotSerializer(many=True, read_only=True)

    class Meta:  # noqa: D106
        model = Fleet
        fields: ClassVar[list[str]] = [
            "uuid",
            "name",
            "created_at",
            "statistics",
            "wallet",
            "bots",
            "exchange_account",
            "testing",
        ]
        read_only_fields: ClassVar[list[str]] = fields

    def __init__(
        self,
        *args: list[any],
        space: Space | None = None,
        **kwargs: dict[str, any],
    ) -> None:
        """."""
        self.space = space
        super().__init__(*args, **kwargs)

    def get_statistics(self, instance: Fleet) -> dict[str, str | int | float]:
        """Return Fleet's stats."""
        return instance.get_stats()

    def get_wallet(self, instance: Fleet) -> list[dict[str, float | str]] | None:
        """Return an aggregated wallet for a given space.

        Return None if space is not given.
        """

        # Method not tested, high chance of being buggy
        def _search_ticker(ticker: str, merged_wallet: list[dict[str, str | float]]) -> int | None:
            """Return the index of the currency in the list if found, None otherwise."""
            for i, currency in enumerate(merged_wallet):
                if currency.get("ticker").ticker == ticker:
                    return i
            return None

        def _update_merged_wallet(index: int, currency: str, merged_wallet: list[dict[str, str | float]]) -> None:
            """Update the merged wallet with the new currency."""
            if index is None:
                merged_wallet.append(
                    {
                        "ticker": currency.ticker,
                        "amount": currency.amount,
                        "mbp": currency.mbp,
                    },
                )
            else:
                merged_wallet[index]["amount"] += currency.amount

        if self.space is None:
            return None
        wallets = ConnectionWallet.objects.filter(owner__owner=self.space.wallet, owner__bot__in=instance.bots)
        merged_wallet: list[dict[str, str | float]] = []

        for wallet in wallets:
            for currency in wallet.currencies.all():
                index = _search_ticker(currency.ticker, merged_wallet)
                _update_merged_wallet(index, currency, merged_wallet)

        return merged_wallet

    def to_representation(self, instance: Fleet | None = None) -> dict[str, any]:
        """Return Fleet's serialization."""
        data = super().to_representation(instance)
        if self.space is not None:
            data["space"] = self.space.uuid
        return data


class FleetMoneyFlowSerializer(serializers.Serializer):
    """."""

    amount = serializers.FloatField(write_only=True, required=True)
    ticker = serializers.CharField(write_only=True, required=True)

    def __init__(
        self,
        side: str,
        *args: list[any],
        space: Space | None = None,
        **kwargs: dict[str, any],
    ) -> None:
        """."""
        self.side = side
        self.space = space
        super().__init__(*args, **kwargs)

    def _invest_validate(self, attrs: dict[str, any]) -> dict[str, any]:
        if self.space.testing:
            space_wallet = self.space.wallet
            try:
                currency: SpaceWallet = space_wallet.currencies.get(ticker=attrs["ticker"])
            except SpaceWallet.DoesNotExist:
                error_msg: str = f"{attrs['ticker']} does not exist in space ({self.space.name})."
                raise serializers.ValidationError(error_msg) from None

            if currency.amount < attrs["amount"]:
                error_msg: str = f"Not enough {currency.ticker} in the wallet."
                raise serializers.ValidationError(error_msg)

            return attrs

        error_msg: str = "Real invest is not implemented yet."
        raise NotImplementedError(error_msg)

    def _withdraw_validate(self, attrs: dict[str, any]) -> dict[str, any]:  # noqa: ARG002
        if self.space.testing:
            error_msg: str = "Withdraw is not implemented yet."
            raise NotImplementedError(error_msg)

        error_msg: str = "Real withdraw is not implemented yet."
        raise NotImplementedError(error_msg)

    def validate(self, attrs: dict[str, any]) -> dict[str, any]:
        """Check if the wallet has enough money to invest."""
        match self.side.upper():
            case "INVEST":
                return self._invest_validate(attrs)
            case "WITHDRAW":
                return self._withdraw_validate(attrs)
            case _:
                error_msg: str = "Invalid side."
                raise ValueError(error_msg)

    def save(self) -> None:
        """Make the transaction."""
        amount = self.validated_data["amount"]
        ticker = self.validated_data["ticker"]
        self.instance.invest(self.space, amount, ticker)



from typing import ClassVar

from rest_framework import serializers

from django_napse.api.bots.serializers.strategy_serializer import StrategySerializer
from django_napse.core.models.bots.plugin import Plugin


class PluginSerializer(serializers.ModelSerializer):
    """Serialize a Plugin instance."""

    strategy = StrategySerializer()

    class Meta:  # noqa: D106
        model = Plugin
        fields = "__all__"
        read_only_fields: ClassVar[list[str]] = [
            "id",
        ]



# Generated by Django 4.2.6 on 2023-10-14 21:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_auth", "0003_alter_keypermission_key"),
    ]

    operations = [
        migrations.AddField(
            model_name="napseapikey",
            name="description",
            field=models.TextField(blank=True),
        ),
    ]



# Generated by Django 4.2.7 on 2024-01-11 22:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0008_bothistory"),
    ]

    operations = [
        migrations.RenameField(
            model_name="order",
            old_name="exit_base_amount",
            new_name="exit_amount_base",
        ),
        migrations.RenameField(
            model_name="order",
            old_name="exit_quote_amount",
            new_name="exit_amount_quote",
        ),
    ]



from typing import ClassVar

from rest_framework import serializers

from django_napse.core.models.bots.config import BotConfig


class ConfigSerializer(serializers.ModelSerializer):
    """Serialize a BotConfig instance."""

    class Meta:  # noqa: D106
        model = BotConfig
        fields = "__all__"
        read_only_fields: ClassVar[list[str]] = [
            "uuid",
            "immutable",
        ]



from typing import TYPE_CHECKING, Literal, Union

from django.db import models

from django_napse.core.models.bots.managers import ArchitectureManager
from django_napse.core.models.wallets.currency import CurrencyPydantic
from django_napse.utils.constants import ORDER_LEEWAY_PERCENTAGE, PLUGIN_CATEGORIES, SIDES
from django_napse.utils.errors.orders import OrderError
from django_napse.utils.findable_class import FindableClass

if TYPE_CHECKING:
    from django_napse.core.models.bots.controller import Controller
    from django_napse.core.models.bots.plugin import Plugin
    from django_napse.core.models.bots.strategy import Strategy
    from django_napse.core.models.connections.connection import Connection

DBDataType = dict[
    Literal[
        "strategy",
        "config",
        "architecture",
        "controllers",
        "connections",
        "connection_data",
        "plugins",
    ],
    Union[
        "Strategy",
        dict[str, any],
        "Architecture",
        dict[str, "Controller"],
        list["Connection"],
        dict["Connection", dict[str, any]],
        dict[str, list["Plugin"]],
    ],
]


class Architecture(models.Model, FindableClass):
    """Define how the bot will get data."""

    objects = ArchitectureManager()

    def __str__(self) -> str:
        return f"ARCHITECHTURE {self.pk}"

    @property
    def variables(self) -> dict[str, any]:
        """Returns variables of the architecture."""
        self = self.find(self.pk)
        variables = {}
        for variable in self._meta.get_fields():
            if variable.name.startswith("variable_"):
                variables[variable.name[8:]] = getattr(self, variable.name)
        return variables

    def get_candles(self):  # pragma: no cover  # noqa: ANN201, D102
        if self.__class__ == Architecture:
            error_msg = "get_candles not implemented for the Architecture base class, please implement it in a subclass."
        else:
            error_msg = f"get_candles not implemented for the Architecture base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def copy(self):  # pragma: no cover # noqa: ANN201, D102
        if self.__class__ == Architecture:
            error_msg = "copy not implemented for the Architecture base class, please implement it in a subclass."
        else:
            error_msg = f"copy not implemented for the Architecture base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def controllers_dict(self):  # pragma: no cover # noqa: ANN201, D102
        if self.__class__ == Architecture:
            error_msg = "controllers_dict not implemented for the Architecture base class, please implement it in a subclass."
        else:
            error_msg = f"controllers_dict not implemented for the Architecture base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def accepted_tickers(self):  # pragma: no cover # noqa: ANN201, D102
        if self.__class__ == Architecture:
            error_msg = "accepted_tickers not implemented for the Architecture base class, please implement it in a subclass."
        else:
            error_msg = f"accepted_tickers not implemented for the Architecture base class, please implement it in the {self.__class__} class."
        raise NotImplementedError(error_msg)

    def accepted_investment_tickers(self):  # pragma: no cover  # noqa: ANN201, D102
        if self.__class__ == Architecture:
            error_msg = "accepted_investment_tickers not implemented for the Architecture base class, please implement it in a subclass."
        else:
            error_msg = (
                f"accepted_investment_tickers not implemented for the Architecture base class, please implement it in the {self.__class__} class."
            )
        raise NotImplementedError(error_msg)

    def get_extras(self):  # noqa
        return {}

    def skip(self, data: dict) -> bool:  # noqa
        return False

    def strategy_modifications(self, order: dict, data) -> list[dict]:  # noqa
        """Return modifications."""
        return []

    def connection_modifications(self, order: dict, data) -> list[dict]:  # noqa
        """Return modifications."""
        return []

    def architecture_modifications(self, order: dict, data) -> list[dict]:  # noqa
        """Return modifications."""
        return []

    def prepare_data(self) -> dict[str, dict[str, any]]:
        """Return candles data."""
        return {
            "candles": {controller: self.get_candles(controller) for controller in self.controllers_dict().values()},
            "extras": self.get_extras(),
        }

    def prepare_db_data(
        self,
    ) -> DBDataType:
        """Return the data that is needed to give orders."""
        return {
            "strategy": self.strategy.find(),
            "config": self.strategy.find().config.find().settings,
            "architecture": self.find(),
            "controllers": self.controllers_dict(),
            "connections": self.strategy.bot.get_connections(),
            "connection_data": self.strategy.bot.get_connection_data(),
            "plugins": {category: self.strategy.plugins.filter(category=category) for category in PLUGIN_CATEGORIES},
        }

    def _get_orders(self, data: dict, no_db_data: DBDataType = None) -> list[dict]:
        data = data or self.prepare_data()
        no_db_data = no_db_data or self.prepare_db_data()
        strategy = no_db_data["strategy"]
        connections = no_db_data["connections"]
        architecture = no_db_data["architecture"]
        all_orders = []
        for connection in connections:
            new_data = {**data, **no_db_data, "connection": connection}
            if architecture.skip(data=new_data):
                continue
            for plugin in no_db_data["plugins"][PLUGIN_CATEGORIES.PRE_ORDER]:
                plugin.apply(data=new_data)
            orders = strategy.give_order(data=new_data)
            for order in orders:
                for plugin in no_db_data["plugins"][PLUGIN_CATEGORIES.POST_ORDER]:
                    plugin.apply(data={**new_data, "order": order})
                order["StrategyModifications"] += architecture.strategy_modifications(order=order, data=new_data)
                order["ConnectionModifications"] += architecture.connection_modifications(order=order, data=new_data)
                order["ArchitectureModifications"] += architecture.architecture_modifications(order=order, data=new_data)
            required_amount = {}
            for order in orders:
                required_amount[order["asked_for_ticker"]] = required_amount.get(order["asked_for_ticker"], 0) + order["asked_for_amount"]

                if order["side"] == SIDES.KEEP and order["asked_for_amount"] != 0:
                    error_msg = f"Order on {order['pair']} has a side of KEEP but an amount of {order['asked_for_amount']}."
                    raise OrderError.ProcessError(error_msg)
                if order["side"] != SIDES.KEEP and order["asked_for_amount"] == 0:
                    error_msg = f"Order on {order['pair']} has a side of {order['side']} but an amount of 0."
                    raise OrderError.ProcessError(error_msg)
            for ticker, amount in required_amount.items():
                if amount > no_db_data["connection_data"][connection]["wallet"].currencies.get(
                    ticker,
                    CurrencyPydantic(ticker=ticker, amount=0, mbp=0),
                ).amount / (1 + (ORDER_LEEWAY_PERCENTAGE + 1) / 100):
                    available = no_db_data["connection_data"][connection]["wallet"].currencies.get(
                        ticker,
                        CurrencyPydantic(ticker=ticker, amount=0, mbp=0),
                    ).amount / (1 + (ORDER_LEEWAY_PERCENTAGE + 1) / 100)
                    for order in [_order for _order in orders if _order["asked_for_ticker"] == ticker]:
                        order["asked_for_amount"] *= available / amount
            all_orders += orders
        return all_orders



import json

from django.apps import apps
from django.db.models.signals import post_migrate
from django.db.transaction import atomic
from django.dispatch import receiver

from django_napse.core.settings import napse_settings as settings


@receiver(post_migrate)
def create_exchanges(sender, **kwargs):
    Exchange = apps.get_model("django_napse_core", "Exchange")
    exchange_configs = settings.NAPSE_EXCHANGE_CONFIGS
    created_exchanges = []
    with atomic():
        for exchange_name, exchange_config in exchange_configs.items():
            _, created = Exchange.objects.get_or_create(name=exchange_name, description=exchange_config["description"])
            if created:
                created_exchanges.append(exchange_name)
    if len(created_exchanges) > 0:
        print(f"Created exchanges: {', '.join(created_exchanges)}")


@receiver(post_migrate)
def create_accounts(sender, **kwargs):
    Exchange = apps.get_model("django_napse_core", "Exchange")
    from django_napse.core.models.accounts.exchange import EXCHANGE_ACCOUNT_DICT

    try:
        with open(settings.NAPSE_SECRETS_FILE_PATH, "r") as json_file:
            secrets = json.load(json_file)
    except FileNotFoundError:
        with open(settings.NAPSE_SECRETS_FILE_PATH, "w") as json_file:
            json.dump({"Exchange Accounts": {}}, json_file)
        secrets = {"Exchange Accounts": {}}
    created_exchange_accounts = []

    with atomic():
        for exchange_id, exchange_secrets in secrets["Exchange Accounts"].items():
            exchange_name = exchange_secrets.pop("exchange")
            exchange = Exchange.objects.get(name=exchange_name)

            try:
                EXCHANGE_ACCOUNT_DICT[exchange_name].objects.get(
                    name=exchange_id,
                    exchange=exchange,
                    **exchange_secrets,
                )
            except EXCHANGE_ACCOUNT_DICT[exchange_name].DoesNotExist:
                EXCHANGE_ACCOUNT_DICT[exchange_name].objects.create(
                    name=exchange_id,
                    exchange=exchange,
                    default=True,
                    **exchange_secrets,
                )
                created_exchange_accounts.append(exchange_id)

    if len(created_exchange_accounts) > 0:
        print(f"Created exchange accounts: {', '.join(created_exchange_accounts)}")



from typing import ClassVar

from django_napse.core.models import EmptyBotConfig, Space
from django_napse.utils.errors import BotConfigError
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_bot_config -v2 --keepdb --parallel
"""


class BotConfigDefaultTestCase:
    def simple_create(self):
        return self.model.objects.create(space=self.space, settings=self.settings)

    def test_error_duplicate(self):
        self.simple_create()
        with self.assertRaises(BotConfigError.DuplicateBotConfig):
            self.simple_create()

    def test_missing_setting(self):
        with self.assertRaises(BotConfigError.MissingSettingError):
            self.model.objects.create(space=self.space, settings={})

    def test_duplicate_immutable(self):
        config = self.simple_create()
        config.duplicate_immutable()
        config.duplicate_immutable()

    def test_duplicate_other_space(self):
        config = self.simple_create()
        with self.assertRaises(ValueError):
            config.duplicate_other_space(self.space)
        new_space = Space.objects.create(name="Test Space 2", exchange_account=self.exchange_account, description="This is a test space")
        config.duplicate_other_space(new_space)


class EmptyBotConfigTestCase(BotConfigDefaultTestCase):
    model = EmptyBotConfig
    settings: ClassVar = {"empty": True}


class EmptyBotConfigBINANCETestCase(EmptyBotConfigTestCase, ModelTestCase):
    exchange = "BINANCE"
